import time
from copy import deepcopy
from typing import Dict, List, Optional

from tabulate import tabulate

import spyctl.api as api
import spyctl.cli as cli
import spyctl.config.configs as cfg
import spyctl.resources.notification_config_templates as nct
import spyctl.spyctl_lib as lib

NOTIFICATIONS_HEADERS = [
    "NAME",
    "ID",
    "TYPE",
    "TARGET",
    "STATUS",
    "AGE",
]

NOTIF_CONFIG_TMPL_HEADERS = [
    "NAME",
    "ID",
    "TYPE",
    "SCHEMA_TYPE",
    "DESCRIPTION",
]

DEFAULT_NOTIFICATION_CONFIG = {
    lib.API_FIELD: lib.API_VERSION,
    lib.KIND_FIELD: lib.NOTIFICATION_KIND,
    lib.METADATA_FIELD: {
        lib.METADATA_TYPE_FIELD: "object",
        lib.NAME_FIELD: None,
    },
    lib.SPEC_FIELD: {
        lib.ENABLED_FIELD: True,
        lib.NOTIF_DEFAULT_SCHEMA: "",
        lib.NOTIF_CONDITION_FIELD: "",
        lib.NOTIF_TITLE_FIELD: "",
        lib.NOTIF_MESSAGE_FIELD: "",
        lib.NOTIF_TEMPLATE_FIELD: "CUSTOM",
        lib.NOTIF_TARGET_FIELD: "",
        lib.NOTIF_ADDITIONAL_FIELDS: {},
    },
}


class NotificationConfig:
    """
    apiVersion: spyderbat/v1
    kind: SpyderbatNotification
    metadata:
      type: object
      name: Agent Health
    spec:
      enabled: true
      schemaType: event_opsflag:agent_offline
      condition: ephemeral = true
      title: "Agent Offline"
      message: "Detected {{ ref.id }} offline at {{ time }}."
      target: TARGET_NAME
      template: TEMPLATE_NAME
      additionalFields:
        slack_icon: ":large_green_circle:"
    """

    # Default config type is "object" but custom type can
    # be determined by schema type such as "metrics"
    config_types = {"event_metric": lib.NOTIF_TYPE_METRICS}

    def __init__(self, config_resource: Dict = None) -> None:
        if not config_resource:
            config_resource = deepcopy(DEFAULT_NOTIFICATION_CONFIG)
        meta: Dict = config_resource[lib.METADATA_FIELD]
        spec: Dict = config_resource[lib.SPEC_FIELD]
        if lib.METADATA_UID_FIELD not in meta:
            self.id = "notif:" + lib.make_uuid()
            meta[lib.METADATA_UID_FIELD] = self.id
            self.new = True
            self.create_time = None
            self.last_updated = None
        else:
            self.id = meta[lib.METADATA_UID_FIELD]
            self.new = False
            self.create_time = meta.get(lib.NOTIF_CREATE_TIME, None)
            self.last_updated = meta.get(lib.NOTIF_LAST_UPDATED, None)
        self.changed = False
        self.name = meta.get(lib.NAME_FIELD, "")
        self.enabled = spec.get(lib.ENABLED_FIELD, True)
        self.schema_type = spec.get(lib.NOTIF_DEFAULT_SCHEMA, "")
        self.sub_schema = spec.get(lib.NOTIF_SUB_SCHEMA)
        self.condition = spec.get(lib.NOTIF_CONDITION_FIELD, "")
        self.title = spec.get(lib.NOTIF_TITLE_FIELD)
        self.message = spec.get(lib.NOTIF_MESSAGE_FIELD)
        self.target = spec.get(lib.NOTIF_TARGET_FIELD, "")
        self.template = spec.get(lib.NOTIF_TEMPLATE_FIELD, "")
        self.additional_fields = spec.get(lib.NOTIF_ADDITIONAL_FIELDS, {})
        self.cooldown = spec.get(lib.NOTIF_COOLDOWN_FIELD)
        self.for_duration = spec.get(lib.NOTIF_FOR_DURATION_FIELD)

    def update(self, **kwargs) -> None:
        for key, value in kwargs.items():
            if hasattr(self, key) and getattr(self, key) != value:
                self.changed = True
                setattr(self, key, value)

    def set_last_updated(self):
        now = time.time()
        self.last_updated = now
        if not self.create_time:
            self.create_time = now

    def as_dict(self) -> Dict:
        rv = {
            lib.API_FIELD: lib.API_VERSION,
            lib.KIND_FIELD: lib.NOTIFICATION_KIND,
            lib.METADATA_FIELD: {
                lib.METADATA_UID_FIELD: self.id,
                lib.METADATA_NAME_FIELD: self.name,
                lib.METADATA_TYPE_FIELD: self.config_type(self.schema_type),
            },
            lib.SPEC_FIELD: {
                lib.ENABLED_FIELD: self.enabled,
                lib.NOTIF_DEFAULT_SCHEMA: self.schema_type,
                lib.NOTIF_CONDITION_FIELD: self.condition,
                lib.NOTIF_TARGET_FIELD: self.target,
                lib.NOTIF_TITLE_FIELD: self.title,
                lib.NOTIF_MESSAGE_FIELD: self.message,
                lib.NOTIF_TEMPLATE_FIELD: self.template,
                lib.NOTIF_ADDITIONAL_FIELDS: self.additional_fields,
            },
        }
        if self.for_duration is not None:
            rv[lib.SPEC_FIELD][
                lib.NOTIF_FOR_DURATION_FIELD
            ] = self.for_duration
        if self.cooldown is not None:
            rv[lib.SPEC_FIELD][lib.NOTIF_COOLDOWN_FIELD] = self.cooldown
        if self.sub_schema is not None:
            rv[lib.SPEC_FIELD][lib.NOTIF_SUB_SCHEMA] = self.sub_schema
        if self.create_time:
            rv[lib.METADATA_FIELD][lib.NOTIF_CREATE_TIME] = self.create_time
            rv[lib.METADATA_FIELD][lib.NOTIF_LAST_UPDATED] = self.last_updated
        return rv

    @property
    def route(self) -> Dict:
        rv = {}
        if isinstance(self.target, str):
            targets = [self.target]
        else:
            targets = self.target
        rv[lib.TARGETS_FIELD] = targets
        rv[lib.DATA_FIELD] = {
            lib.NOTIF_CREATE_TIME: self.create_time,
            lib.ID_FIELD: self.id,
            lib.NOTIF_LAST_UPDATED: self.last_updated,
            lib.NOTIF_SETTINGS_FIELD: self.as_dict(),
            lib.NOTIF_NAME_FIELD: self.name,
        }
        rv[lib.ROUTE_EXPR] = {"property": "data.route_id", "equals": self.id}
        return rv

    @property
    def type(self) -> str:
        return self.config_type(self.schema_type)

    @classmethod
    def config_type(cls, schema_type: str):
        if schema_type in cls.config_types:
            return cls.config_types[schema_type]
        else:
            return lib.NOTIF_TYPE_OBJECT


class NotificationConfigTemplate:
    def __init__(self, config_template: Dict) -> None:
        self.display_name = config_template["display_name"]
        self.description = config_template["description"]
        self.config_values = config_template["config"]
        self.id = config_template["id"]
        self.type = config_template["type"]

    def update_rt(self, rt: NotificationConfig):
        for key in rt.settings[lib.SPEC_FIELD]:
            if key in self.pre_conf_fields:
                rt.update(**{key: self.pre_conf_fields[key]})

    def as_dict(self):
        rv = {
            lib.API_FIELD: lib.API_VERSION,
            lib.KIND_FIELD: lib.NOTIF_TMPL_KIND,
            lib.METADATA_FIELD: {
                lib.METADATA_NAME_FIELD: self.display_name,
                lib.METADATA_UID_FIELD: self.id,
                lib.METADATA_TYPE_FIELD: self.type,
            },
            lib.SPEC_FIELD: {
                lib.TMPL_DESCRIPTION_FIELD: self.description,
                lib.TMPL_CONFIG_VALUES_FIELD: self.config_values,
            },
        }
        return rv


def __load_notif_configs() -> List[NotificationConfigTemplate]:
    rv = []
    for config in nct.TEMPLATES:
        rv.append(NotificationConfigTemplate(config))
    return rv


NOTIF_CONFIG_TEMPLATES: List[
    NotificationConfigTemplate
] = __load_notif_configs()


def notif_config_tmpl_summary_output(
    templates: List[NotificationConfigTemplate],
):
    data = []
    for tmpl in templates:
        data.append(
            [
                tmpl.display_name,
                tmpl.id,
                tmpl.type,
                tmpl.config_values["schema_type"],
                __wrap_text(tmpl.description, 45),
            ]
        )
    data.sort(key=lambda row: row[2])
    return tabulate(data, NOTIF_CONFIG_TMPL_HEADERS, "plain")


def __wrap_text(text, max_line_length=30) -> str:
    lines = text.split("\n")
    wrapped_lines = []

    for line in lines:
        while len(line) > max_line_length:
            # Find the last space within the max_line_length
            last_space = line.rfind(" ", 0, max_line_length)
            if last_space == -1:
                # If no space is found, split the line at max_line_length
                last_space = max_line_length

            # Append the portion of the line up to the last space to the result
            wrapped_lines.append(line[:last_space])
            # Remove the portion that was added to the result
            last_space_plus_one = last_space + 1
            line = line[last_space_plus_one:]

        # Append the remaining portion of the line to the result
        wrapped_lines.append(line)

    return "\n".join(wrapped_lines)


def notifications_summary_output(routes: Dict, notif_type: str):
    data = []
    if (
        notif_type == lib.NOTIF_TYPE_ALL
        or notif_type == lib.NOTIF_TYPE_DASHBOARD
    ):
        dashboard_search_notifications = __parse_legacy_notifications(routes)
        if dashboard_search_notifications:
            data.extend(__get_dashboard_data(dashboard_search_notifications))
    notif_configs = __parse_notification_configs(routes)
    if notif_type != lib.NOTIF_TYPE_ALL:
        notif_configs = list(
            filter(lambda cfg: cfg.type == notif_type, notif_configs)
        )
    data.extend(__get_config_data(notif_configs))
    data.sort(key=lambda row: (row[2], row[0]))
    return tabulate(data, NOTIFICATIONS_HEADERS, "plain")


def notifications_wide_output(routes: Dict, notif_type: str):
    data = []
    if (
        notif_type == lib.NOTIF_TYPE_ALL
        or notif_type == lib.NOTIF_TYPE_DASHBOARD
    ):
        dashboard_search_notifications = __parse_legacy_notifications(routes)
        if dashboard_search_notifications:
            data.extend(
                __get_dashboard_data_wide(dashboard_search_notifications)
            )
    notif_configs = __parse_notification_configs(routes)
    if notif_type != lib.NOTIF_TYPE_ALL:
        notif_configs = list(
            filter(lambda cfg: cfg.type == notif_type, notif_configs)
        )
    data.extend(__get_config_data_wide(notif_configs))
    data.sort(key=lambda row: (row[2], row[0]))
    return tabulate(data, NOTIFICATIONS_HEADERS, "plain")


def __parse_notification_configs(routes: Dict) -> List[NotificationConfig]:
    rv = []
    for route in routes:
        if __is_notification_config(route):
            config = NotificationConfig(
                route[lib.DATA_FIELD][lib.NOTIF_SETTINGS_FIELD]
            )
            rv.append(config)
    return rv


def __parse_legacy_notifications(routes: Dict):
    rv = []
    if not isinstance(routes, list):
        return rv
    for route in routes:
        if __is_dashboard_notification(route):
            rv.append(route)
    return rv


def __is_notification_config(route: Dict) -> bool:
    if not isinstance(route, dict):
        return False
    data: Dict = route.get(lib.NOTIF_DATA_FIELD)
    if not isinstance(data, dict):
        return False
    settings = data.get(lib.NOTIF_SETTINGS_FIELD)
    if not isinstance(settings, dict):
        return False
    kind = settings.get(lib.KIND_FIELD)
    if kind == lib.NOTIFICATION_KIND:
        return True
    return False


def __is_dashboard_notification(route: Dict) -> bool:
    if not isinstance(route, dict):
        return False
    data: Dict = route.get(lib.NOTIF_DATA_FIELD)
    if not isinstance(data, dict):
        # This is the default notification type and data is
        # an optional field managed by the UI
        return True
    settings = data.get(lib.NOTIF_SETTINGS_FIELD)
    if not isinstance(settings, dict):
        # Same reason as above
        return True
    kind = settings.get(lib.KIND_FIELD)
    if kind == lib.NOTIFICATION_KIND:
        # In case we get explicit about dashboard search notifications
        notif_type = lib.get_metadata_type(settings)
        if notif_type == lib.NOTIF_TYPE_DASHBOARD:
            return True
        else:
            return False
    return True


def __get_config_data(configs: List[NotificationConfig]):
    table_rows = []
    for config in configs:
        if isinstance(config.target, list):
            if len(config.target) == 1:
                target = config.target[0]
            else:
                target = f"{len(config.target)} targets"
        else:
            target = config.target
        status = "Enabled" if config.enabled else "Disabled"
        table_rows.append(
            [
                config.name,
                config.id,
                config.type,
                target,
                status,
                lib.calc_age(config.create_time),
            ]
        )
    return table_rows


def __get_config_data_wide(configs: List[NotificationConfig]):
    table_rows = []
    for config in configs:
        if isinstance(config.target, list):
            if len(config.target) == 1:
                target = config.target[0]
            else:
                target = "\n".join(config.target)
        else:
            target = config.target
        status = "Enabled" if config.enabled else "Disabled"
        table_rows.append(
            [
                config.name,
                config.id,
                config.type,
                target,
                status,
                lib.calc_age(config.create_time),
            ]
        )
    return table_rows


def __get_dashboard_data(routes: List[Dict]):
    table_rows = []
    for route in routes:
        targets = route.get(lib.TARGETS_FIELD)
        if targets:
            if len(targets) == 1:
                target = targets[0]
            else:
                target = f"{len(targets)} targets"
        else:
            target = lib.NOT_AVAILABLE
        data = route.get(lib.DATA_FIELD, {})
        if isinstance(data, dict):
            id = data.get(lib.ID_FIELD, lib.NOT_AVAILABLE)
            name = data.get(lib.NAME_FIELD, lib.NOT_AVAILABLE)
            create_time = data.get(lib.NOTIF_CREATE_TIME)
            if create_time:
                age = lib.calc_age(create_time)
            else:
                age = lib.NOT_AVAILABLE
        else:
            id = lib.NOT_AVAILABLE
            name = lib.NOT_AVAILABLE
            age = lib.NOT_AVAILABLE
        table_rows.append(
            [
                name,
                id,
                lib.NOTIF_TYPE_DASHBOARD,
                target,
                "Enabled",
                age,
            ]
        )
    return table_rows


def __get_dashboard_data_wide(routes: List[Dict]):
    table_rows = []
    for route in routes:
        targets = route.get(lib.TARGETS_FIELD)
        if targets:
            if len(targets) == 1:
                target = targets[0]
            else:
                target = "\n".join(targets)
        else:
            target = lib.NOT_AVAILABLE
        data = route.get(lib.DATA_FIELD, {})
        if isinstance(data, dict):
            id = data.get(lib.ID_FIELD, lib.NOT_AVAILABLE)
            name = data.get(lib.NAME_FIELD, lib.NOT_AVAILABLE)
            create_time = data.get(lib.NOTIF_CREATE_TIME)
            if create_time:
                age = lib.calc_age(create_time)
            else:
                age = lib.NOT_AVAILABLE
        else:
            id = lib.NOT_AVAILABLE
            name = lib.NOT_AVAILABLE
            age = lib.NOT_AVAILABLE
        table_rows.append(
            [
                name,
                id,
                lib.NOTIF_TYPE_DASHBOARD,
                target,
                "Enabled",
                age,
            ]
        )
    return table_rows


def create_config(name, tgt_name_or_id, tmpl_name_or_id):
    import spyctl.resources.notification_targets as nt

    config = NotificationConfig()
    target: nt.NotificationTarget = nt.get_target(tgt_name_or_id)
    if not target:
        cli.err_exit(f'No target with name or uid "{tgt_name_or_id}"')
    if tmpl_name_or_id != "CUSTOM":
        template = get_template(tmpl_name_or_id)
    else:
        template = tmpl_name_or_id
    if not template:
        cli.err_exit(f'No template with name or uid "{tmpl_name_or_id}"')
    if template == "CUSTOM":
        config.update(template="CUSTOM")
    else:
        config.update(template=template.display_name)
        config.update(**template.config_values)
    config.update(target=target.name)
    config.update(name=name)
    return config.as_dict()


def get_template(name_or_id) -> Optional[NotificationConfigTemplate]:
    for tmpl in NOTIF_CONFIG_TEMPLATES:
        if name_or_id == tmpl.display_name or name_or_id == tmpl.id:
            return tmpl


def put_and_get_notif_pol(
    nr: NotificationConfig = None, delete_id: str = None
):
    ctx = cfg.get_current_context()
    n_pol = api.get_notification_policy(*ctx.get_api_data())
    routes: List = n_pol.get(lib.ROUTES_FIELD, [])
    if delete_id:
        for i, route in list(enumerate(routes)):
            rt_id = route.get(lib.DATA_FIELD, {}).get(lib.ID_FIELD)
            if not rt_id:
                continue
            if rt_id == delete_id:
                routes.pop(i)
                break
    if nr:
        found = False
        for i, route in list(enumerate(routes)):
            rt_id = route.get(lib.DATA_FIELD, {}).get(lib.ID_FIELD)
            if not rt_id:
                continue
            if rt_id == nr.id:
                found = True
                routes[i] = nr.route
                break
        if not found:
            routes.append(nr.route)
    n_pol[lib.ROUTES_FIELD] = routes
    api.put_notification_policy(*ctx.get_api_data(), n_pol)
    n_pol = api.get_notification_policy(*ctx.get_api_data())
    return n_pol



import traceback

from fastapi import HTTPException


def bad_request(msg: str = None):
    detail = "Bad Request."
    if msg:
        detail += f" {msg}"
    raise HTTPException(400, detail=detail)


def internal_server_error(msg: str = None):
    s = traceback.format_exc()
    detail = "Internal Server Error." + s + "\n"
    if msg:
        detail += f" {msg}"
    raise ValueError(detail)



import spyctl.api as api
import spyctl.cli as cli
import spyctl.config.configs as cfg
import spyctl.spyctl_lib as lib
import spyctl.resources.notification_targets as nt


def handle_test_notification(test_targets=[]):
    if not test_targets:
        cli.err_exit("No targets provided.")
    ctx = cfg.get_current_context()
    notif_pol = api.get_notification_policy(*ctx.get_api_data())
    if not notif_pol or not notif_pol.get(lib.TARGETS_FIELD):
        cli.err_exit(
            "No targets to test. Use 'spyctl create notification-target'."
        )
    pol_targets = notif_pol.get(lib.TARGETS_FIELD)
    if not pol_targets:
        cli.err_exit("No targets to test.")
    for name_or_id in test_targets:
        test_target = None
        # check if name exists
        if name_or_id in pol_targets:
            tgt_data = pol_targets[name_or_id]
            test_target = nt.Target(backend_target={name_or_id: tgt_data})
        if not test_target:
            for tgt_name, tgt in pol_targets.items():
                id = tgt.get(lib.DATA_FIELD, {}).get(lib.ID_FIELD)
                if id is None:
                    continue
                if id == name_or_id:
                    test_target = nt.Target(backend_target={tgt_name: tgt})
                    break
        if not test_target:
            cli.err_exit(f"No notification targets matching '{name_or_id}'.")
        resp = api.post_test_notification(
            *ctx.get_api_data(), test_target.name
        )
        if resp.status_code == 200:
            cli.try_log(f"Successfully sent test to '{test_target.name}'")



import spyctl.spyctl_lib as lib
from typing import IO
import json


def handle_print_file(file: IO, list_output: bool):
    if list_output:
        data = [lib.load_file_for_api_test(file)]
    else:
        data = lib.load_file_for_api_test(file)
    data = json.dumps({"data": json.dumps(data)})
    print(data)



import json

from tabulate import tabulate

# step 1: spyctl get connections -o json --ndjson --remote-port 53 > connections53.json and
# spyctl get containers -o json --ndjson > containers53.json


# load connection data and add it to the conn dictionary
# if any field is not present default is " "
conn_dict = {}
with open("connections53.json", "r") as file:
    for connections in file:
        connection = json.loads(connections)
        containerUID = connection.get("container_uid", " ")
        payload = connection.get("payload", " ")
        conn_dict[containerUID] = {"payload": payload}


container_dict = {}

# load container data and add it to the container dict with
# containerUID as the key
with open("containers53.json", "r") as file:
    for containers in file:
        containerInfo = json.loads(containers)
        containerUID = containerInfo.get("id", " ")
        containerName = containerInfo.get("container_name_k8s", " ")
        # print(containerName)
        cluster = containerInfo.get("clustername", " ")
        pod = containerInfo.get("pod_name", " ")
        namespace = containerInfo.get("pod_namespace", " ")

        container_dict[containerUID] = {
            "containerUID": containerUID,
            "cluster": cluster,
            "pod": pod,
            "namespace": namespace,
        }

# match the container--> connection data based on connection key.
matched_data = []
for connection_key, connection_data in conn_dict.items():
    if connection_key in container_dict:
        matched_data.append(
            {
                "containerUID": connection_key,
                "container_k8s": containerName,
                "pod": pod,
                "cluster": cluster,
                "namespace": namespace,
                "payload": conn_dict[connection_key]["payload"],
            }
        )

# print the data in tabulate manner.
headers = ["containerID", "container_k8s", "pod", "cluster", "namespace", "payload"]
table = tabulate(
    [list(d.values()) for d in matched_data], headers=headers, tablefmt="plain"
)

print(table)



from typing import Dict, List, Tuple

from tabulate import tabulate

import spyctl.api as api
import spyctl.config.configs as cfg
import spyctl.spyctl_lib as lib

SUMMARY_HEADERS = [
    "IMAGE",
    "IMAGE_ID",
    "LATEST_TIMESTAMP",
    "COUNT",
    "NAMESPACE",
    "CLUSTER",
]


class ContainerGroup:
    def __init__(self) -> None:
        self.latest_timestamp = lib.NOT_AVAILABLE
        self.count = 0
        self.image = None
        self.image_id = None
        self.namespace = None
        self.cluster = None

    def add_container(self, cont: Dict):
        self.__update_latest_timestamp(cont.get("time"))
        self.count += 1
        if not self.image:
            self.image = cont["image"]
            self.image_id = cont["image_id"]
            self.namespace = cont.get("pod_namespace")
            self.cluster = cont.get("clustername") or cont.get("cluster_uid")

    def summary_data(self) -> List[str]:
        rv = [
            self.image,
            self.image_id,
            lib.epoch_to_zulu(self.latest_timestamp),
            self.count,
            self.namespace or lib.NOT_AVAILABLE,
            self.cluster or lib.NOT_AVAILABLE,
        ]
        return rv

    def __update_latest_timestamp(self, timestamp):
        if timestamp is None:
            return
        if self.latest_timestamp == lib.NOT_AVAILABLE:
            self.latest_timestamp = timestamp
        elif timestamp > self.latest_timestamp:
            self.latest_timestamp = timestamp


def cont_summary_output(
    ctx: cfg.Context,
    muids: List[str],
    time: Tuple[float, float],
    pipeline=None,
    limit_mem=False,
):
    cont_groups: Dict[str, ContainerGroup] = {}
    for container in api.get_containers(
        *ctx.get_api_data(),
        muids,
        time,
        pipeline=pipeline,
        limit_mem=limit_mem,
    ):
        key = __key(container)
        if key not in cont_groups:
            cont_groups[key] = ContainerGroup()
        cont_groups[key].add_container(container)
    data = []
    for group in cont_groups.values():
        data.append(group.summary_data())
    data.sort(key=lambda x: (x[5], x[0], x[1], x[4]))
    rv = tabulate(
        data,
        headers=SUMMARY_HEADERS,
        tablefmt="plain",
    )
    return rv


def cont_summary_data(container: Dict) -> List[str]:
    return [
        container[lib.BE_CONTAINER_IMAGE],
        container[lib.BE_CONTAINER_IMAGE_ID],
        container[lib.STATUS_FIELD],
        lib.calc_age(container[lib.VALID_FROM_FIELD]),
    ]


def __key(cont: Dict):
    return (
        cont["image"],
        cont["image_id"],
        cont.get("pod_namespace"),
        cont.get("cluster_name") or cont.get("cluster_uid"),
    )



import os
import shutil
from fnmatch import fnmatch
from pathlib import Path

from click.testing import CliRunner

from spyctl import spyctl
from spyctl.config.configs import CURR_CONTEXT_NONE, set_testing

environ = dict(os.environ)
API_KEY = os.environ.get("API_KEY")
API_URL = os.environ.get("API_URL")
ORG = os.environ.get("ORG")
TEST_SECRET = "__test_secret__"
TEST_CONTEXT = "__test_context__"
TEST_CONTEXT2 = "__test_context2__"
CURR_PATH = Path.cwd()
WORKSPACE_DIR_NAME = "test_workspace__"
WORKSPACE_PATH = Path(f"./{WORKSPACE_DIR_NAME}")
WORKSPACE_CONFIG = Path(str(WORKSPACE_PATH.absolute()) + "/.spyctl/config")
CURRENT_CONTEXT = None


class SetupException(Exception):
    pass


def env_setup() -> bool:
    # Tests that the proper environment variables exist in pyproject.toml
    if not API_KEY or API_KEY == "__NONE__":
        print(
            "No api key provided. Edit the API_KEY environment variable in"
            " pyproject.toml"
        )
        return False
    if not API_URL or API_URL == "__NONE__":
        print(
            "No api url provided. Edit the API_URL environment variable in"
            ' pyproject.toml to "API_URL=https://api.spyderbat.com"'
        )
        return False
    if not ORG or ORG == "__NONE__":
        print(
            "No organization provided. Edit the ORG environment variable in"
            " pyproject.toml"
        )
        return False
    return True


def test_secrets():
    runner = CliRunner()
    test_key = "test_key"
    test_url = "https://test.url"
    # Test set a new secret
    result = runner.invoke(
        spyctl.main,
        [
            "config",
            "set-apisecret",
            "-k",
            API_KEY,
            "-u",
            API_URL,
            TEST_SECRET,
        ],
    )
    assert f"Set new apisecret '{TEST_SECRET}' in" in result.output
    # Test getting all secrets
    result = runner.invoke(
        spyctl.main,
        [
            "config",
            "get-apisecrets",
        ],
    )
    assert TEST_SECRET in result.output
    # Test getting single secret
    result = runner.invoke(
        spyctl.main,
        [
            "config",
            "get-apisecrets",
            TEST_SECRET,
        ],
    )
    assert TEST_SECRET in result.output
    # Test getting single secret in yaml output
    result = runner.invoke(
        spyctl.main,
        [
            "config",
            "get-apisecrets",
            "-o",
            "yaml",
            TEST_SECRET,
        ],
    )
    assert TEST_SECRET in result.output
    assert API_KEY in result.output
    assert API_URL in result.output
    # Test updating a secret
    result = runner.invoke(
        spyctl.main,
        [
            "config",
            "set-apisecret",
            "-k",
            test_key,
            "-u",
            test_url,
            TEST_SECRET,
        ],
    )
    assert f"Updated apisecret '{TEST_SECRET}' in" in result.output
    result = runner.invoke(
        spyctl.main,
        [
            "config",
            "get-apisecrets",
            "-o",
            "yaml",
            TEST_SECRET,
        ],
    )
    assert TEST_SECRET in result.output
    assert test_key in result.output
    assert test_url in result.output
    # Test deleting a secret
    result = runner.invoke(
        spyctl.main,
        [
            "config",
            "delete-apisecret",
            "-y",
            TEST_SECRET,
        ],
    )
    assert f"Deleted secret '{TEST_SECRET}' from" in result.output
    result = runner.invoke(
        spyctl.main,
        [
            "config",
            "get-apisecrets",
        ],
    )
    assert TEST_SECRET not in result.output


def test_contexts():
    runner = CliRunner()
    create_secret()
    # Create a new context
    result = runner.invoke(
        spyctl.main,
        [
            "config",
            "set-context",
            "-s",
            TEST_SECRET,
            "-o",
            ORG,
            "-u",
            TEST_CONTEXT,
        ],
    )
    assert (
        f"Set new context '{TEST_CONTEXT}' in configuration file"
        in result.output
    )
    result = runner.invoke(
        spyctl.main,
        [
            "config",
            "set-context",
            "-s",
            TEST_SECRET,
            "-o",
            ORG,
            TEST_CONTEXT2,
        ],
    )
    # Test get all contexts
    test_str = f"*{TEST_CONTEXT}*{ORG}*0*"
    result = runner.invoke(
        spyctl.main,
        [
            "config",
            "get-contexts",
        ],
    )
    lines = result.output.splitlines()
    match = False
    for line in lines:
        if fnmatch(line, test_str):
            match = True
            assert "*" in line
            break
    assert match
    # Get specific contexts
    result = runner.invoke(
        spyctl.main,
        ["config", "get-contexts", TEST_CONTEXT],
    )
    assert fnmatch(result.output, test_str)
    # Display current context
    result = runner.invoke(
        spyctl.main,
        ["config", "current-context"],
    )
    assert result.output.strip("\n") == TEST_CONTEXT
    # Test use-context (switching contexts)
    result = runner.invoke(
        spyctl.main,
        ["config", "use-context", TEST_CONTEXT2],
    )
    assert f"Set current context to '{TEST_CONTEXT2}' in" in result.output
    result = runner.invoke(
        spyctl.main,
        ["config", "current-context"],
    )
    assert result.output.strip("\n") == TEST_CONTEXT2
    # Display config file
    result = runner.invoke(
        spyctl.main,
        ["config", "view"],
    )
    assert f"- name: {TEST_CONTEXT}" in result.output
    assert f"- name: {TEST_CONTEXT2}" in result.output
    # Delete contexts
    result = runner.invoke(
        spyctl.main,
        ["config", "delete-context", TEST_CONTEXT],
    )
    assert f"Deleted context '{TEST_CONTEXT}' in" in result.output
    result = runner.invoke(
        spyctl.main,
        ["config", "delete-context", TEST_CONTEXT2],
    )
    assert f"Deleted context '{TEST_CONTEXT2}' in" in result.output
    result = runner.invoke(
        spyctl.main,
        ["config", "get-contexts"],
    )
    assert TEST_CONTEXT not in result.output
    assert TEST_CONTEXT2 not in result.output


def test_workspace():
    create_secret()
    runner = CliRunner()
    # Test initialize a workspace
    result = runner.invoke(
        spyctl.main,
        [
            "config",
            "init-workspace",
        ],
    )
    test_str = (
        f"Created configuration file at */{WORKSPACE_DIR_NAME}/.spyctl/config*"
    )
    assert fnmatch(result.output, test_str)
    result = runner.invoke(
        spyctl.main,
        [
            "config",
            "set-context",
            "-s",
            TEST_SECRET,
            "-o",
            ORG,
            TEST_CONTEXT,
        ],
    )
    assert result.exit_code == 0
    result = runner.invoke(
        spyctl.main,
        [
            "config",
            "set-context",
            "-s",
            TEST_SECRET,
            "-o",
            ORG,
            TEST_CONTEXT2,
        ],
    )
    assert result.exit_code == 0
    # Create global contexts with the same name for testing purposes
    create_context(TEST_CONTEXT)
    create_context(TEST_CONTEXT2)
    # Get context from merged config
    result = runner.invoke(
        spyctl.main,
        [
            "config",
            "get-contexts",
            "-o",
            "wide",
            TEST_CONTEXT,
        ],
    )
    test_str = f"*/{WORKSPACE_DIR_NAME}/.spyctl/config*"
    assert TEST_CONTEXT in result.output
    assert fnmatch(result.output, test_str)
    # Delete global context by same name
    result = runner.invoke(
        spyctl.main,
        [
            "config",
            "delete-context",
            "-y",
            "-g",
            TEST_CONTEXT,
        ],
    )
    result = runner.invoke(
        spyctl.main,
        [
            "config",
            "get-contexts",
            "-g",
        ],
    )
    assert TEST_CONTEXT not in result.output
    result = runner.invoke(
        spyctl.main,
        [
            "config",
            "get-contexts",
        ],
    )
    assert TEST_CONTEXT in result.output
    # Delete workspace context
    result = runner.invoke(
        spyctl.main,
        [
            "config",
            "get-contexts",
            "-w",
        ],
    )
    assert TEST_CONTEXT2 in result.output
    result = runner.invoke(
        spyctl.main,
        [
            "config",
            "delete-context",
            "-y",
            TEST_CONTEXT2,
        ],
    )
    result = runner.invoke(
        spyctl.main,
        [
            "config",
            "get-contexts",
            "-w",
        ],
    )
    assert TEST_CONTEXT2 not in result.output
    # Set workspace current context to context within the global config
    result = runner.invoke(
        spyctl.main,
        [
            "config",
            "use-context",
            "-g",
            TEST_CONTEXT2,
        ],
    )
    test_str = (
        f"Set current context to '{TEST_CONTEXT2}' in configuration file"
    )
    assert test_str in result.output
    assert WORKSPACE_DIR_NAME not in result.output
    result = runner.invoke(
        spyctl.main,
        ["config", "current-context", "-g"],
    )
    assert TEST_CONTEXT2 == result.output.strip("\n")
    # Set workspace current context to context within the workspace config
    result = runner.invoke(
        spyctl.main,
        [
            "config",
            "use-context",
            TEST_CONTEXT,
        ],
    )
    test_str = f"Set current context to '{TEST_CONTEXT}' in configuration file"
    assert test_str in result.output
    assert WORKSPACE_DIR_NAME in result.output
    result = runner.invoke(
        spyctl.main,
        [
            "config",
            "current-context",
        ],
    )
    assert TEST_CONTEXT == result.output.strip("\n")
    # Test ensure unable to set global current context to context that
    # exists only in the local workspace config.
    result = runner.invoke(
        spyctl.main,
        [
            "config",
            "use-context",
            "-g",
            TEST_CONTEXT,
        ],
    )
    test_str = f"Unable to set current context '{TEST_CONTEXT}' for"
    assert test_str in result.output


def create_secret():
    runner = CliRunner()
    result = runner.invoke(
        spyctl.main,
        [
            "config",
            "set-apisecret",
            "-k",
            API_KEY,
            "-u",
            API_URL,
            TEST_SECRET,
        ],
    )
    if result.exit_code != 0:
        raise SetupException("Unable to create test secret")


def delete_secret():
    runner = CliRunner()
    runner.invoke(
        spyctl.main,
        [
            "config",
            "delete-apisecret",
            "-y",
            TEST_SECRET,
        ],
    )


def create_context(name):
    runner = CliRunner()
    runner.invoke(
        spyctl.main,
        [
            "config",
            "set-context",
            "-s",
            TEST_SECRET,
            "-o",
            ORG,
            "-g",
            name,
        ],
    )


def delete_context(name):
    runner = CliRunner()
    runner.invoke(
        spyctl.main,
        ["config", "delete-context", name],
    )


def current_context():
    runner = CliRunner()
    current_ctx = runner.invoke(
        spyctl.main,
        [
            "config",
            "current-context",
        ],
    ).output.strip("\n")
    return current_ctx


def use_current_context():
    if CURRENT_CONTEXT:
        runner = CliRunner()
        runner.invoke(spyctl.main, ["config", "use-context", CURRENT_CONTEXT])


def setup_module():
    if not env_setup():
        raise SetupException("Check environment variables in pyproject.toml")
    global CURRENT_CONTEXT
    current_ctx = current_context()
    if current_ctx and current_ctx != CURR_CONTEXT_NONE:
        CURRENT_CONTEXT = current_ctx
    if WORKSPACE_PATH.exists():
        teardown_module()
        raise SetupException("Workspace path already exists")
    WORKSPACE_PATH.mkdir(exist_ok=False)
    os.chdir(WORKSPACE_PATH)
    # Set a global variable in configs.py to force reloading
    # the config file. Otherwise the global variable LOADED_CONFIG
    # remains set across invocations
    set_testing()


def teardown_module():
    delete_context(TEST_CONTEXT)
    delete_context(TEST_CONTEXT2)
    delete_secret()
    use_current_context()
    os.chdir(CURR_PATH)
    shutil.rmtree(WORKSPACE_PATH, ignore_errors=True)



import ipaddress
from typing import Dict, List, Tuple

import zulu
from tabulate import tabulate

import spyctl.config.configs as cfg
import spyctl.spyctl_lib as lib
import spyctl.api as api

NOT_AVAILABLE = lib.NOT_AVAILABLE

SUMMARY_HEADERS = [
    "DESTINATION_IP",
    # "DESTINATION_PORT",
    "DIRECTION",
    "PROCESS_NAME",
    "COUNT",
    "LATEST_TIMESTAMP",
]


class ConnectionGroup:
    def __init__(self) -> None:
        self.ref_conn = None
        self.latest_timestamp = NOT_AVAILABLE
        self.count = 0
        self.ip = None

    def add_conn(self, conn: Dict):
        self.__update_latest_timestamp(conn.get("time"))
        if self.ref_conn is None:
            self.ref_conn = conn
        self.count += 1
        ip = ipaddress.ip_address(conn["remote_ip"])
        if self.ip is None:
            self.ip = ip.exploded
        else:
            self.ip = _loose_abbrev_ips(self.ip, ip.exploded)

    def __update_latest_timestamp(self, timestamp):
        if timestamp is None:
            return
        if self.latest_timestamp == NOT_AVAILABLE:
            self.latest_timestamp = timestamp
        elif timestamp > self.latest_timestamp:
            self.latest_timestamp = timestamp

    def summary_data(self, ignore_ips) -> List[str]:
        timestamp = NOT_AVAILABLE
        if self.latest_timestamp != NOT_AVAILABLE:
            timestamp = str(
                zulu.Zulu.fromtimestamp(self.latest_timestamp).format(
                    "YYYY-MM-ddTHH:mm:ss"
                )
            )
        rv = [
            # self.ref_conn["remote_ip"],
            # self.ref_conn["remote_port"],
            _shorten_v6(self.ip),
            self.ref_conn["direction"],
            self.ref_conn["proc_name"],
            str(self.count),
            timestamp,
        ]
        if ignore_ips:
            rv = rv[1:]
        return rv


def conn_summary_output(
    ctx: cfg.Context,
    muids: List[str],
    time: Tuple[float, float],
    ignore_ips=False,
    pipeline=None,
    limit_mem=False,
) -> str:
    groups: Dict[str, ConnectionGroup] = {}
    for conn in api.get_connections(
        *ctx.get_api_data(),
        muids,
        time,
        limit_mem=limit_mem,
        pipeline=pipeline,
    ):
        key = _key(conn, ignore_ips)
        if key not in groups:
            groups[key] = ConnectionGroup()
        groups[key].add_conn(conn)
    data = []
    for group in groups.values():
        data.append(group.summary_data(ignore_ips))
    sort_key = (
        (lambda x: [x[0], x[1]])
        if ignore_ips
        else (lambda x: [x[1], x[0], x[2]])
    )
    output = tabulate(
        sorted(data, key=sort_key),
        headers=SUMMARY_HEADERS,
        tablefmt="plain",
    )
    return output


def _key(connection: Dict, ignore_ips):
    if ignore_ips:
        return (
            connection["direction"],
            connection["proc_name"],
        )
    ip = ipaddress.ip_address(connection["remote_ip"])
    # just like the unnecessary complexity of this summary
    ip_str = ip.exploded
    found = 0
    for i, char in enumerate(ip_str):
        if not char.isdigit():
            found += 1
            if found == 2:
                ip_str = ip_str[: i + 1]
                break
    return (
        ip.version,
        ip_str,
        connection["direction"],
        connection["proc_name"],
    )


def _loose_abbrev_ips(ip1, ip2):
    if ip1 == ip2:
        return ip1
    ret = ""
    for char1, char2 in zip(ip1, ip2):
        if char1 == char2:
            ret += char1
        else:
            ret += "*"
            return ret
    return ret + "*"


def _shorten_v6(ip):
    if ":" not in ip:
        return ip
    ip = ip.replace("0000", "zero")
    i = 0
    pos = -1
    most = 0
    last = 0
    num = 0
    while i < len(ip):
        if ip[i : i + 4] == "zero":
            if num == 0:
                last = i
            num += 1
            if num > most:
                most = num
                pos = last
        else:
            num = 0
        i += 5
    len_diff = 39 - len(ip)
    if ip.endswith("*"):
        len_diff += 1
    if pos != -1:
        ip = ip[:pos] + ":" + ip[pos + most * 5 :]
        if pos == 0:
            ip = ":" + ip
    ip = ip.replace("0", "")
    ip = ip.replace("zero", "0")
    if len_diff > 0:
        ip += f" (+{len_diff})"
    return ip


def connections_output(conns: List[Dict]) -> Dict:
    if len(conns) == 1:
        return conns[0]
    elif len(conns) > 1:
        return {
            lib.API_FIELD: lib.API_VERSION,
            lib.ITEMS_FIELD: conns,
        }
    else:
        return {}



from typing import Dict, List, Tuple

from tabulate import tabulate

import spyctl.api as api
import spyctl.config.configs as cfg
import spyctl.spyctl_lib as lib

SUMMARY_HEADERs = [
    "NAME",
    "READY",
    "LAST_STATUS_SEEN",
    "LAST_SEEN",
    "RESTARTS",
    "AGE",
    "NAMESPACE",
    "CLUSTER",
]


def pods_output_summary(
    ctx: cfg.Context,
    clusters: List[str],
    time: Tuple[float, float],
    pipeline=None,
    limit_mem=False,
) -> str:
    data = []
    for pod in api.get_pods(
        *ctx.get_api_data(), clusters, time, pipeline, limit_mem
    ):
        data.append(__pod_summary_data(pod))
    output = tabulate(
        sorted(data, key=lambda x: [x[7], x[6], x[0]]),
        headers=SUMMARY_HEADERs,
        tablefmt="plain",
    )
    return output


def __pod_summary_data(pod: Dict) -> List:
    k8s_status = pod[lib.BE_K8S_STATUS]
    container_statuses = k8s_status.get(lib.CONTAINER_STATUSES_FIELD, {})
    ready = __get_ready_count(container_statuses)
    restarts = __calc_restarts(container_statuses)
    meta = pod[lib.METADATA_FIELD]
    rv = [
        meta[lib.METADATA_NAME_FIELD],
        ready,
        k8s_status[lib.BE_PHASE],
        lib.epoch_to_zulu(pod["time"]),
        restarts,
        lib.calc_age(lib.to_timestamp(meta[lib.METADATA_CREATE_TIME])),
        meta[lib.NAMESPACE_FIELD],
        pod.get("cluster_name") or pod.get("cluster_uid"),
    ]
    return rv


def __get_ready_count(container_statuses: Dict):
    count = 0
    ready_count = 0
    for cont in container_statuses:
        count += 1
        if cont.get("ready", False):
            ready_count += 1
    return f"{ready_count}/{count}"


def __calc_restarts(container_statuses: Dict):
    count = 0
    for cont in container_statuses:
        count += cont.get("restartCount", 0)
    return count



"""Handles generation of filters for use by the API instead of doing local
filtering.
"""

from copy import deepcopy
from typing import Dict, List, Tuple, Union

import spyctl.api as api
import spyctl.config.configs as cfg
import spyctl.filter_resource as filt
import spyctl.spyctl_lib as lib

# Source types
SOURCE_TYPE_CLUID = "cluid"
SOURCE_TYPE_CLUID_BASE = "cluid_base"
SOURCE_TYPE_CLUID_CBUN = "cluid_cbun"
SOURCE_TYPE_CLUID_POCO = "cluid_poco"
SOURCE_TYPE_CLUID_FLAG = "cluid_flag"
CLUSTER_SOURCES = [
    SOURCE_TYPE_CLUID,
    SOURCE_TYPE_CLUID_BASE,
    SOURCE_TYPE_CLUID_CBUN,
    SOURCE_TYPE_CLUID_FLAG,
    SOURCE_TYPE_CLUID_POCO,
]
SOURCE_TYPE_GLOBAL = "global"
SOURCE_TYPE_MUID = "muid"
SOURCE_TYPE_POL = "pol"

POLICIES_CACHE = None


def get_filtered_cluids(**filters) -> List[str]:
    ctx = cfg.get_current_context()
    clusters = api.get_clusters(*ctx.get_api_data())
    clusters = filt.filter_clusters(clusters, **filters)
    cluids = [c["uid"] for c in clusters]
    return cluids


def get_filtered_muids(**filters) -> List[str]:
    ctx = cfg.get_current_context()
    sources = api.get_sources(*ctx.get_api_data())
    sources = filt.filter_sources(sources, **filters)
    muids = [s["uid"] for s in sources]
    return muids


def get_filtered_pol_uids(**filters) -> List[str]:
    global POLICIES_CACHE
    ctx = cfg.get_current_context()
    if not POLICIES_CACHE:
        policies = api.get_policies(*ctx.get_api_data())
        POLICIES_CACHE = policies
    else:
        policies = POLICIES_CACHE
    policy_filters = filters.get(lib.POLICIES_FIELD)
    if policy_filters:
        policies = [
            p
            for p in filt.filter_obj(
                policies,
                [
                    [lib.METADATA_FIELD, lib.METADATA_NAME_FIELD],
                    [lib.METADATA_FIELD, lib.METADATA_UID_FIELD],
                ],
                policy_filters,
            )
        ]
    policy_uids = [
        p[lib.METADATA_FIELD][lib.METADATA_UID_FIELD] for p in policies
    ]
    return policy_uids


class API_Filter:
    property_map = (
        {}
    )  # property -> field name on object (. notation for nested fields)
    name_or_uid_props: []  # properties in the property_map that are related to name_or_id filtering
    values_helper = {
        lib.CLUSTER_FIELD: get_filtered_cluids,
        lib.MACHINE_SELECTOR_FIELD: get_filtered_muids,
    }  # property -> callable that takes the filter value and turns it into something more useable. Ex. cluster name to cluster uid
    source_type = SOURCE_TYPE_MUID
    alternate_source_type = None

    @classmethod
    def generate_pipeline(
        cls,
        schema,
        name_or_uid=None,
        latest_model=True,
        filters={},
        count=False,
    ) -> List:
        pipeline_items = []
        if latest_model:
            pipeline_items.append({"latest_model": {}})
        pipeline_items.append(
            cls.__generate_fprint_api_filters(schema, name_or_uid, **filters)
        )
        if count:
            pipeline_items.append(
                {
                    "aggregation": {
                        "aggregations": [{"count": {}, "as": "count"}]
                    }
                }
            )
        return pipeline_items

    @classmethod
    def __generate_fprint_api_filters(
        cls,
        schema,
        name_or_uid: Union[str, List],
        **filters,
    ) -> Dict:
        and_items = [{"schema": schema}]
        if name_or_uid:
            and_items.append(
                cls.__build_or_block(
                    cls.name_or_uid_props,
                    [name_or_uid],
                )
            )
        for key, values in filters.items():
            if isinstance(values, list) and len(values) > 1:
                and_items.append(cls.__build_or_block([key], values))
            else:
                if isinstance(values, list):
                    value = values[0]
                else:
                    value = values
                property = cls.__build_property(key)
                if not property:
                    continue
                if isinstance(value, int) or isinstance(value, float):
                    and_items.append({"property": property, "equals": value})
                elif "*" in value or "?" in value:
                    value = lib.simple_glob_to_regex(value)
                    and_items.append({"property": property, "re_match": value})
                else:
                    and_items.append({"property": property, "equals": value})
        if len(and_items) > 1:
            rv = {"filter": {"and": and_items}}
        else:
            rv = {"filter": and_items[0]}
        return rv

    @classmethod
    def __build_or_block(cls, keys: List[str], values: List[str]):
        or_items = []
        for key in keys:
            for value in values:
                if "*" in value or "?" in value:
                    value = lib.simple_glob_to_regex(value)
                    or_items.append(
                        {
                            "property": cls.__build_property(key),
                            "re_match": value,
                        }
                    )
                else:
                    or_items.append(
                        {
                            "property": cls.__build_property(key),
                            "equals": value,
                        }
                    )
        return {"or": or_items}

    @classmethod
    def __build_property(cls, key: str):
        return cls.property_map[key]

    @classmethod
    def build_sources_and_filters(cls, **filters) -> Tuple[List[str], Dict]:
        ctx = cfg.get_current_context()
        ctx_filters = deepcopy(ctx.get_filters())
        sources = cls.__get_sources(ctx_filters, filters)
        filters = cls.__get_filters(ctx_filters, filters)
        return sources, filters

    @classmethod
    def __get_sources(cls, ctx_filters: Dict, filters: Dict):
        ctx = cfg.get_current_context()
        sources = []
        if cls.source_type == SOURCE_TYPE_GLOBAL:
            sources.append(ctx.global_source)
            if cls.alternate_source_type == SOURCE_TYPE_MUID and (
                lib.MACHINES_FIELD in ctx_filters
                or lib.MACHINES_FIELD in filters
            ):
                muids = get_filtered_muids(**filters)
                cls.__pop_muid_filters(ctx_filters, filters)
                if muids:
                    sources = muids
        elif cls.source_type == SOURCE_TYPE_CLUID:
            cluids = get_filtered_cluids(**filters)
            cls.__pop_cluid_filters(ctx_filters, filters)
            sources = cluids
        elif cls.source_type == SOURCE_TYPE_CLUID_BASE:
            cluids = get_filtered_cluids(**filters)
            cls.__pop_cluid_filters(ctx_filters, filters)
            sources = [cluid + "_base" for cluid in cluids]
        elif cls.source_type == SOURCE_TYPE_CLUID_POCO:
            cluids = get_filtered_cluids(**filters)
            cls.__pop_cluid_filters(ctx_filters, filters)
            sources = [cluid + "_poco" for cluid in cluids]
        elif cls.source_type == SOURCE_TYPE_POL:
            pol_uids = get_filtered_pol_uids(**filters)
            cls.__pop_pol_filters(ctx_filters, filters)
            sources = pol_uids
        else:  # muids is the default
            if cls.alternate_source_type in CLUSTER_SOURCES and (
                lib.CLUSTER_FIELD in ctx_filters
                or lib.CLUSTER_FIELD in filters
            ):
                cluids = get_filtered_cluids(**filters)
                cls.__pop_cluid_filters(ctx_filters, filters)
                if cls.alternate_source_type == SOURCE_TYPE_CLUID_CBUN:
                    sources = [cluid + "_cbun" for cluid in cluids]
                elif cls.alternate_source_type == SOURCE_TYPE_CLUID_FLAG:
                    sources = [cluid + "_flag" for cluid in cluids]
                elif cls.alternate_source_type == SOURCE_TYPE_CLUID_POCO:
                    sources = [cluid + "_poco" for cluid in cluids]
                else:
                    sources = cluids
            else:
                muids = get_filtered_muids(**filters)
                cls.__pop_muid_filters(ctx_filters, filters)
                sources = muids
        return sources

    @classmethod
    def __get_filters(cls, ctx_filters: Dict, filters: Dict):
        rv_filters = {}
        for key, value in ctx_filters.items():
            if key in cls.property_map:
                rv_filters[key] = value
        for key, value in filters.items():
            if key in cls.property_map:
                rv_filters[key] = value
        for key, value in rv_filters.items():
            if key in cls.values_helper:
                value = cls.values_helper[key](**{key: value})
                rv_filters[key] = value
        return rv_filters

    @staticmethod
    def __pop_cluid_filters(ctx_filters: Dict, cmdline_filters: Dict):
        ctx_filters.pop(lib.CLUSTER_FIELD, None)
        cmdline_filters.pop(lib.CLUSTER_FIELD, None)

    @staticmethod
    def __pop_muid_filters(ctx_filters: Dict, cmdline_filters: Dict):
        ctx_filters.pop(lib.MACHINES_FIELD, None)
        cmdline_filters.pop(lib.MACHINES_FIELD, None)

    @staticmethod
    def __pop_pol_filters(ctx_filters: Dict, cmdline_filters: Dict):
        ctx_filters.pop(lib.POLICIES_FIELD, None)
        cmdline_filters.pop(lib.POLICIES_FIELD, None)


class Agents(API_Filter):
    property_map = {
        lib.MACHINES_FIELD: "muid",
        lib.AGENT_ID: "id",
        lib.AGENT_HOSTNAME: "hostname",
        lib.STATUS_FIELD: lib.STATUS_FIELD,
    }
    name_or_uid_props = [lib.AGENT_ID, lib.AGENT_HOSTNAME, lib.MACHINES_FIELD]
    source_type = SOURCE_TYPE_GLOBAL
    alternate_source_type = SOURCE_TYPE_MUID

    @classmethod
    def generate_pipeline(
        cls, name_or_uid=None, type=None, latest_model=True, filters={}
    ) -> List:
        # TODO implement type when supported by analytics
        schema = lib.MODEL_AGENT_SCHEMA_PREFIX
        return super(Agents, cls).generate_pipeline(
            schema, name_or_uid, latest_model, filters
        )


class AgentMetrics(API_Filter):
    property_map = {lib.ID_FIELD: lib.ID_FIELD}
    name_or_uid_props = [lib.ID_FIELD]
    source_type = SOURCE_TYPE_MUID

    @classmethod
    def generate_pipeline(
        cls, name_or_uid=None, type=None, latest_model=True, filters={}
    ) -> List:
        schema = f"{lib.EVENT_METRICS_PREFIX}:{lib.EVENT_METRIC_SUBTYPE_MAP['agent']}"
        return super(AgentMetrics, cls).generate_pipeline(
            schema, name_or_uid, latest_model, filters
        )


class Connections(API_Filter):
    property_map = {
        lib.ID_FIELD: lib.ID_FIELD,
        lib.PROC_NAME_FIELD: lib.PROC_NAME_FIELD,
        lib.REMOTE_HOSTNAME_FIELD: lib.REMOTE_HOSTNAME_FIELD,
        lib.PROTOCOL_FIELD: "proto",
        lib.STATUS_FIELD: lib.STATUS_FIELD,
        lib.LOCAL_PORT: lib.LOCAL_PORT,
        lib.REMOTE_PORT: lib.REMOTE_PORT,
    }
    name_or_uid_props = [
        lib.ID_FIELD,
        lib.REMOTE_HOSTNAME_FIELD,
        lib.PROC_NAME_FIELD,
    ]
    source_type = SOURCE_TYPE_MUID

    @classmethod
    def generate_pipeline(
        cls, name_or_uid=None, latest_model=True, filters={}
    ) -> List:
        schema = lib.MODEL_CONNECTION_PREFIX
        return super(Connections, cls).generate_pipeline(
            schema, name_or_uid, latest_model, filters
        )


class ConnectionBundles(API_Filter):
    property_map = {
        lib.ID_FIELD: lib.ID_FIELD,
        lib.PROTOCOL_FIELD: "proto",
        lib.CLIENT_PORT: lib.CLIENT_PORT,
        lib.SERVER_PORT: lib.SERVER_PORT,
    }
    name_or_uid_props = [
        lib.ID_FIELD,
    ]
    source_type = SOURCE_TYPE_MUID
    alternate_source_type = SOURCE_TYPE_CLUID_CBUN

    @classmethod
    def generate_pipeline(
        cls, name_or_uid=None, latest_model=True, filters={}
    ) -> List:
        schema = lib.MODEL_CONN_BUN_PREFIX
        return super(ConnectionBundles, cls).generate_pipeline(
            schema, name_or_uid, latest_model, filters
        )


class Containers(API_Filter):
    property_map = {
        lib.ID_FIELD: lib.ID_FIELD,
        lib.CONTAINER_ID_FIELD: lib.BE_CONTAINER_ID,
        lib.IMAGE_FIELD: lib.BE_CONTAINER_IMAGE,
        lib.CONTAINER_NAME_FIELD: lib.BE_CONTAINER_NAME,
        lib.IMAGEID_FIELD: lib.BE_CONTAINER_IMAGE_ID,
        lib.STATUS_FIELD: lib.STATUS_FIELD,
        lib.NAMESPACE_FIELD: "pod_namespace",
    }
    name_or_uid_props = [
        lib.ID_FIELD,
        lib.CONTAINER_ID_FIELD,
        lib.IMAGE_FIELD,
        lib.CONTAINER_NAME_FIELD,
        lib.IMAGEID_FIELD,
    ]
    source_type = SOURCE_TYPE_MUID
    alternate_source_type = SOURCE_TYPE_CLUID_POCO

    @classmethod
    def generate_pipeline(
        cls, name_or_uid=None, latest_model=True, filters={}
    ) -> List:
        schema = lib.MODEL_CONTAINER_PREFIX
        return super(Containers, cls).generate_pipeline(
            schema, name_or_uid, latest_model, filters
        )


class Deployments(API_Filter):
    property_map = {
        lib.ID_FIELD: lib.ID_FIELD,
        lib.BE_KUID_FIELD: lib.BE_KUID_FIELD,
        lib.NAME_FIELD: f"{lib.METADATA_FIELD}.{lib.METADATA_NAME_FIELD}",
        lib.NAMESPACE_FIELD: f"{lib.METADATA_FIELD}.{lib.METADATA_NAMESPACE_FIELD}",
    }
    name_or_uid_props = [lib.ID_FIELD, lib.NAME_FIELD, lib.BE_KUID_FIELD]
    source_type = SOURCE_TYPE_CLUID_BASE

    @classmethod
    def generate_pipeline(
        cls, name_or_uid=None, latest_model=True, filters={}
    ) -> List:
        schema = lib.MODEL_DEPLOYMENT_PREFIX
        return super(Deployments, cls).generate_pipeline(
            schema, name_or_uid, latest_model, filters
        )


class Deviations(API_Filter):
    property_map = {
        lib.ID_FIELD: lib.ID_FIELD,
        lib.POLICIES_FIELD: lib.BE_POL_UID_FIELD,
    }
    name_or_uid_props = [lib.ID_FIELD]
    source_type = SOURCE_TYPE_POL

    @classmethod
    def generate_pipeline(
        cls, name_or_uid=None, latest_model=True, filters={}
    ) -> List:
        schema = (
            f"{lib.EVENT_AUDIT_PREFIX}:"
            f"{lib.EVENT_AUDIT_SUBTYPE_MAP['deviation']}"
        )
        return super(Deviations, cls).generate_pipeline(
            schema, name_or_uid, latest_model, filters
        )

    @classmethod
    def generate_count_pipeline(cls, name_or_uid, filters={}):
        schema = (
            f"{lib.EVENT_AUDIT_PREFIX}:"
            f"{lib.EVENT_AUDIT_SUBTYPE_MAP['deviation']}"
        )
        pipeline_items = super(Deviations, cls).generate_pipeline(
            schema,
            name_or_uid,
            True,
            filters,
        )
        pipeline_items.append(
            {
                "aggregation": {
                    "aggregations": [
                        {
                            "uniq_count": {"property": "checksum"},
                            "as": "counts",
                        },
                    ],
                    "by": [{"property": "policy_uid"}],
                },
            }
        )
        return pipeline_items


class Fingerprints(API_Filter):
    property_map = {
        lib.MACHINES_FIELD: "muid",
        lib.POD_FIELD: "pod_uid",
        lib.CLUSTER_FIELD: "cluster_uid",
        lib.NAMESPACE_FIELD: f"{lib.METADATA_FIELD}.{lib.METADATA_NAMESPACE_FIELD}",
        lib.CGROUP_FIELD: "cgroup",
        lib.IMAGE_FIELD: "image",
        lib.IMAGEID_FIELD: "image_id",
        lib.CONTAINER_ID_FIELD: "container_id",
        lib.CONTAINER_NAME_FIELD: "container_name",
        lib.STATUS_FIELD: lib.STATUS_FIELD,
        lib.ID_FIELD: lib.ID_FIELD,
    }
    source_type = SOURCE_TYPE_GLOBAL
    alternate_source_type = SOURCE_TYPE_MUID

    name_or_uid_props = [
        lib.IMAGE_FIELD,
        lib.IMAGEID_FIELD,
        lib.CGROUP_FIELD,
        lib.ID_FIELD,
    ]

    @classmethod
    def generate_pipeline(
        cls, name_or_uid=None, type=None, latest_model=True, filters={}
    ) -> List:
        if type == lib.POL_TYPE_CONT or type == lib.POL_TYPE_SVC:
            schema = (
                f"{lib.MODEL_FINGERPRINT_PREFIX}:"
                f"{lib.MODEL_FINGERPRINT_SUBTYPE_MAP[type]}"
            )
        else:
            schema = f"{lib.MODEL_FINGERPRINT_PREFIX}:"
        return super(Fingerprints, cls).generate_pipeline(
            schema, name_or_uid, latest_model, filters
        )


class Machines(API_Filter):
    property_map = {lib.ID_FIELD: lib.ID_FIELD}
    name_or_uid_props = [lib.ID_FIELD]
    source_type = SOURCE_TYPE_MUID

    @classmethod
    def generate_pipeline(
        cls, name_or_uid=None, latest_model=True, filters={}
    ) -> List:
        schema = lib.MODEL_MACHINE_PREFIX
        return super(Machines, cls).generate_pipeline(
            schema, name_or_uid, latest_model, filters
        )


class Namespaces(API_Filter):
    property_map = {
        lib.ID_FIELD: lib.ID_FIELD,
        lib.BE_KUID_FIELD: f"{lib.METADATA_FIELD}.{lib.METADATA_UID_FIELD}",
        lib.NAME_FIELD: f"{lib.METADATA_FIELD}.{lib.METADATA_NAME_FIELD}",
        lib.NAMESPACE_FIELD: f"{lib.METADATA_FIELD}.{lib.METADATA_NAMESPACE_FIELD}",
    }
    name_or_uid_props = [lib.ID_FIELD, lib.NAME_FIELD, lib.BE_KUID_FIELD]
    source_type = SOURCE_TYPE_CLUID_BASE

    @classmethod
    def generate_pipeline(
        cls, name_or_uid=None, latest_model=True, filters={}
    ) -> List:
        schema = lib.MODEL_NAMESPACE_PREFIX
        return super(Namespaces, cls).generate_pipeline(
            schema, name_or_uid, latest_model, filters
        )

    @classmethod
    def get_name_or_uid_fields(cls):
        rv = []
        for prop in cls.name_or_uid_props:
            rv.append(cls.property_map[prop])
        return rv


class Nodes(API_Filter):
    property_map = {
        lib.ID_FIELD: lib.ID_FIELD,
        lib.BE_KUID_FIELD: lib.BE_KUID_FIELD,
        lib.NAME_FIELD: f"{lib.METADATA_FIELD}.{lib.METADATA_NAME_FIELD}",
        lib.NAMESPACE_FIELD: f"{lib.METADATA_FIELD}.{lib.METADATA_NAMESPACE_FIELD}",
    }
    name_or_uid_props = [lib.ID_FIELD, lib.NAME_FIELD, lib.BE_KUID_FIELD]
    source_type = SOURCE_TYPE_CLUID_BASE

    @classmethod
    def generate_pipeline(
        cls, name_or_uid=None, latest_model=True, filters={}
    ) -> List:
        schema = lib.MODEL_NODE_PREFIX
        return super(Nodes, cls).generate_pipeline(
            schema, name_or_uid, latest_model, filters
        )


class OpsFlags(API_Filter):
    property_map = {lib.ID_FIELD: lib.ID_FIELD}
    name_or_uid_props = [lib.ID_FIELD]
    source_type = SOURCE_TYPE_MUID
    alternate_source_type = SOURCE_TYPE_CLUID_FLAG

    @classmethod
    def generate_pipeline(
        cls, name_or_uid=None, latest_model=True, filters={}
    ) -> List:
        schema = lib.EVENT_OPSFLAG_PREFIX
        return super(OpsFlags, cls).generate_pipeline(
            schema, name_or_uid, latest_model, filters
        )


class Pods(API_Filter):
    property_map = {
        lib.ID_FIELD: lib.ID_FIELD,
        lib.BE_KUID_FIELD: lib.BE_KUID_FIELD,
        lib.NAME_FIELD: f"{lib.METADATA_FIELD}.{lib.METADATA_NAME_FIELD}",
        lib.NAMESPACE_FIELD: f"{lib.METADATA_FIELD}.{lib.METADATA_NAMESPACE_FIELD}",
    }
    name_or_uid_props = [lib.ID_FIELD, lib.METADATA_NAME_FIELD]
    source_type = SOURCE_TYPE_CLUID_POCO

    @classmethod
    def generate_pipeline(
        cls, name_or_uid=None, latest_model=True, filters={}
    ) -> List:
        schema = lib.MODEL_POD_PREFIX
        return super(Pods, cls).generate_pipeline(
            schema, name_or_uid, latest_model, filters
        )


class Processes(API_Filter):
    property_map = {
        lib.ID_FIELD: lib.ID_FIELD,
        lib.CONTAINER_ID_FIELD: "container",
        "user": "euser",
        lib.CGROUP_FIELD: lib.CGROUP_FIELD,
        lib.EXE_FIELD: lib.EXE_FIELD,
        lib.NAME_FIELD: lib.NAME_FIELD,
    }
    name_or_uid_props = [
        lib.NAME_FIELD,
        lib.ID_FIELD,
    ]
    source_type = SOURCE_TYPE_MUID

    @classmethod
    def generate_pipeline(
        cls, name_or_uid=None, latest_model=True, filters={}
    ) -> List:
        schema = lib.MODEL_PROCESS_PREFIX
        return super(Processes, cls).generate_pipeline(
            schema, name_or_uid, latest_model, filters
        )


class RedFlags(API_Filter):
    property_map = {
        lib.ID_FIELD: lib.ID_FIELD,
        "short_name": "short_name",
        lib.CLUSTER_FIELD: "cluster_uid",
    }
    name_or_uid_props = [lib.ID_FIELD, "short_name"]
    source_type = SOURCE_TYPE_MUID
    alternate_source_type = SOURCE_TYPE_CLUID_FLAG

    @classmethod
    def generate_pipeline(
        cls, name_or_uid=None, latest_model=True, filters={}
    ) -> List:
        schema = lib.EVENT_REDFLAG_PREFIX
        return super(RedFlags, cls).generate_pipeline(
            schema, name_or_uid, latest_model, filters
        )


class Spydertraces(API_Filter):
    property_map = {
        lib.ID_FIELD: lib.ID_FIELD,
        lib.MACHINES_FIELD: "muid",
        lib.POD_FIELD: "pod_uid",
        lib.CLUSTER_FIELD: "cluster_uid",
        lib.CGROUP_FIELD: "trigger_cgroup",
        lib.IMAGE_FIELD: "image",
        lib.IMAGEID_FIELD: "image_id",
        lib.CONTAINER_ID_FIELD: "container",
        lib.BE_SCORE: lib.BE_SCORE,
        lib.BE_SUPPRESSED: lib.BE_SUPPRESSED,
        lib.STATUS_FIELD: lib.STATUS_FIELD,
        lib.BE_ROOT_PROC_NAME: lib.BE_ROOT_PROC_NAME,
        lib.BE_TRIGGER_NAME: lib.BE_TRIGGER_NAME,
    }
    name_or_uid_props = [
        lib.BE_ROOT_PROC_NAME,
        lib.BE_TRIGGER_NAME,
        lib.ID_FIELD,
    ]
    source_type = SOURCE_TYPE_MUID

    @classmethod
    def generate_pipeline(
        cls, name_or_uid=None, latest_model=True, filters={}
    ) -> List:
        schema = lib.MODEL_SPYDERTRACE_PREFIX
        return super(Spydertraces, cls).generate_pipeline(
            schema, name_or_uid, latest_model, filters
        )


class SpydertraceSummaries(API_Filter):
    property_map = {lib.ID_FIELD: lib.ID_FIELD}
    name_or_uid_props = [lib.ID_FIELD]
    source_type = SOURCE_TYPE_MUID

    @classmethod
    def generate_pipeline(
        cls, name_or_uid=None, latest_model=True, filters={}
    ) -> List:
        schema = f"{lib.MODEL_FINGERPRINT_PREFIX}:{lib.POL_TYPE_TRACE}"
        return super(SpydertraceSummaries, cls).generate_pipeline(
            schema, name_or_uid, latest_model, filters
        )


class UID_List(API_Filter):
    @classmethod
    def generate_pipeline(cls, uid_list: Dict):
        pipeline_items = [{"latest_model": {}}]
        or_items = []
        for uid in uid_list[lib.DATA_FIELD][lib.UIDS_FIELD]:
            or_items.append({"property": lib.ID_FIELD, "equals": uid})
        filter = {"filter": {"or": or_items}}
        pipeline_items.append(filter)
        return pipeline_items



import json
from typing import Dict, List, Optional, Tuple, Union, Set

import zulu
from tabulate import tabulate

import spyctl.api as api
import spyctl.cli as cli
import spyctl.config.configs as cfg
import spyctl.filter_resource as filt
import spyctl.merge_lib as m_lib
import spyctl.resources.api_filters as _af
import spyctl.resources.fingerprints as spyctl_fprints
import spyctl.resources.resources_lib as r_lib
import spyctl.schemas_v2 as schemas
import spyctl.spyctl_lib as lib

FPRINT_KIND = spyctl_fprints.FPRINT_KIND
GROUP_KIND = spyctl_fprints.GROUP_KIND
BASELINE_KIND = lib.BASELINE_KIND
POLICY_KIND = lib.POL_KIND

POLICY_META_MERGE_SCHEMA = m_lib.MergeSchema(
    lib.METADATA_FIELD,
    merge_functions={
        lib.METADATA_NAME_FIELD: m_lib.keep_base_value_merge,
        lib.METADATA_TYPE_FIELD: m_lib.keep_base_value_merge,
        lib.METADATA_UID_FIELD: m_lib.keep_base_value_merge,
        lib.METADATA_CREATE_TIME: m_lib.keep_base_value_merge,
        lib.LATEST_TIMESTAMP_FIELD: m_lib.greatest_value_merge,
    },
)
POLICY_MERGE_SCHEMAS = [POLICY_META_MERGE_SCHEMA, m_lib.SPEC_MERGE_SCHEMA]


class InvalidPolicyError(Exception):
    pass


class Policy:
    required_keys = {
        lib.API_FIELD,
        lib.KIND_FIELD,
        lib.METADATA_FIELD,
        lib.SPEC_FIELD,
    }

    def __init__(
        self,
        obj: Dict,
        name: str = None,
        mode: str = lib.POL_MODE_AUDIT,
        disable_procs: str = None,
        disable_conns: str = None,
    ) -> None:
        for key in self.required_keys:
            if key not in obj:
                raise InvalidPolicyError(f"Missing {key} for input object")
        self.metadata = obj[lib.METADATA_FIELD]
        if name:
            self.metadata[lib.METADATA_NAME_FIELD] = name
        self.spec: Dict = obj[lib.SPEC_FIELD]
        self.response_actions = obj[lib.SPEC_FIELD].get(
            lib.RESPONSE_FIELD, lib.RESPONSE_ACTION_TEMPLATE
        )
        self.mode = obj[lib.SPEC_FIELD].get(lib.POL_MODE_FIELD, mode)
        self.spec[lib.POL_MODE_FIELD] = self.mode
        self.spec[lib.RESPONSE_FIELD] = self.response_actions
        self.__parse_disable_procs(disable_procs)
        self.__parse_disable_conns(disable_conns)

    def spec_dict(self):
        spec_field_names = [
            lib.PROC_POLICY_FIELD,
            lib.NET_POLICY_FIELD,
            lib.RESPONSE_FIELD,
        ]
        selectors = {}
        other_fields = {}
        pol_fields = {}
        for k, v in self.spec.items():
            if "Selector" in k:
                selectors[k] = v
            if k in spec_field_names:
                continue
            else:
                other_fields[k] = v
        for name in spec_field_names:
            pol_fields[name] = self.spec[name]
        rv = {}
        rv.update(selectors)
        rv.update(other_fields)
        rv.update(pol_fields)
        return rv

    def as_dict(self):
        rv = {
            lib.API_FIELD: lib.API_VERSION,
            lib.KIND_FIELD: POLICY_KIND,
            lib.METADATA_FIELD: self.metadata,
            lib.SPEC_FIELD: self.spec_dict(),
        }
        return rv

    def __parse_disable_procs(self, disable_procs: Optional[str]):
        if disable_procs == lib.DISABLE_PROCS_ALL:
            self.spec[lib.DISABLE_PROCS_FIELD] = lib.DISABLE_PROCS_ALL

    def __parse_disable_conns(self, disable_conns: Optional[str]):
        if disable_conns == lib.DISABLE_CONNS_ALL:
            self.spec[lib.DISABLE_CONNS_FIELD] = lib.DISABLE_CONNS_ALL
        elif disable_conns == lib.DISABLE_CONNS_EGRESS:
            self.spec[lib.DISABLE_CONNS_FIELD] = lib.EGRESS_FIELD
        elif disable_conns == lib.DISABLE_CONNS_INGRESS:
            self.spec[lib.DISABLE_CONNS_FIELD] = lib.INGRESS_FIELD
        elif disable_conns == lib.DISABLE_CONNS_PRIVATE:
            self.spec[lib.DISABLE_PR_CONNS_FIELD] = lib.DISABLE_CONNS_ALL
        elif disable_conns == lib.DISABLE_CONNS_PRIVATE_E:
            self.spec[lib.DISABLE_PR_CONNS_FIELD] = lib.EGRESS_FIELD
        elif disable_conns == lib.DISABLE_CONNS_PRIVATE_I:
            self.spec[lib.DISABLE_PR_CONNS_FIELD] = lib.INGRESS_FIELD
        elif disable_conns == lib.DISABLE_CONNS_PUBLIC:
            self.spec[lib.DISABLE_PU_CONNS_FIELD] = lib.DISABLE_CONNS_ALL
        elif disable_conns == lib.DISABLE_CONNS_PUBLIC_E:
            self.spec[lib.DISABLE_PU_CONNS_FIELD] = lib.EGRESS_FIELD
        elif disable_conns == lib.DISABLE_CONNS_PUBLIC_I:
            self.spec[lib.DISABLE_PU_CONNS_FIELD] = lib.INGRESS_FIELD


def get_data_for_api_call(policy: Policy) -> Tuple[Optional[str], str]:
    policy = policy.as_dict()
    name = policy[lib.METADATA_FIELD][lib.METADATA_NAME_FIELD]
    type = policy[lib.METADATA_FIELD][lib.METADATA_TYPE_FIELD]
    tags = policy[lib.METADATA_FIELD].get(lib.METADATA_TAGS_FIELD)
    uid = policy[lib.METADATA_FIELD].get(lib.METADATA_UID_FIELD, "")
    policy_selectors = {
        key: value
        for key, value in policy[lib.SPEC_FIELD].items()
        if key.endswith("Selector")
    }
    data = {
        lib.API_REQ_FIELD_NAME: name[:32],
        lib.API_REQ_FIELD_POLICY: json.dumps(policy),
        lib.API_REQ_FIELD_POL_SELECTORS: json.dumps(policy_selectors),
        lib.API_REQ_FIELD_TYPE: type,
        lib.API_REQ_FIELD_UID: uid,
    }
    if tags:
        data[lib.API_REQ_FIELD_TAGS] = tags
    else:
        data[lib.API_REQ_FIELD_TAGS] = []
    return uid, data


def create_policy(
    input_data: Union[Dict, List[Dict]],
    mode: str,
    name: str = None,
    ctx: cfg.Context = None,
    disable_procs: str = None,
    disable_conns: str = None,
):
    input_objs = []
    if isinstance(input_data, list):
        if len(input_data) == 0:
            cli.err_exit("Nothing to build policy with")
        for datum in input_data:
            input_objs.extend(r_lib.handle_input_data(datum, ctx))
    else:
        input_objs.extend(r_lib.handle_input_data(input_data, ctx))
    if len(input_objs) == 0:
        cli.err_exit("Nothing to build policy with")
    merge_object = m_lib.MergeObject(
        input_objs[0],
        POLICY_MERGE_SCHEMAS,
        None,
        disable_procs=disable_procs,
        disable_conns=disable_conns,
    )
    if len(input_objs) == 1:
        merge_object.asymmetric_merge({})
    else:
        for obj in input_objs[1:]:
            merge_object.symmetric_merge(obj)
    try:
        policy = Policy(
            merge_object.get_obj_data(),
            name,
            mode,
            disable_procs,
            disable_conns,
        )
    except InvalidPolicyError as e:
        cli.err_exit(f"Unable to create policy. {' '.join(e.args)}")
    # Validate the policy.
    rv = policy.as_dict()
    if not schemas.valid_object(rv):
        cli.err_exit("Created policy failed validation.")
    return rv


def policies_output(policies: List[Dict]):
    if len(policies) == 1:
        return policies[0]
    elif len(policies) > 1:
        return {lib.API_FIELD: lib.API_VERSION, lib.ITEMS_FIELD: policies}
    else:
        return {}


def matching_fingerprints_summary(
    ctx: cfg.Context,
    policies: List[Dict],
    time: Tuple[float, float],
    limit_mem,
):
    has_matching = []
    no_matching = []
    sources, filters = _af.Fingerprints.build_sources_and_filters()
    pipeline = _af.Fingerprints.generate_pipeline(filters=filters)
    fingerprints = list(
        api.get_fingerprints(
            *ctx.get_api_data(),
            sources,
            time,
            pipeline=pipeline,
            limit_mem=limit_mem,
        )
    )
    for policy in policies:
        filters = lib.selectors_to_filters(policy)
        if filt.filter_fingerprints(
            fingerprints,
            use_context_filters=False,
            suppress_warning=True,
            **filters,
        ):
            has_matching.append(policy)
        else:
            no_matching.append(policy)
    output_list = []
    headers = ["UID", "NAME", "STATUS", "TYPE", "CREATE_TIME"]
    if len(no_matching) > 0:
        output_list.append(
            "Policies WITH NO matching fingerprints in last query:"
        )
        no_match_data = []
        for pol in no_matching:
            no_match_data.append(policy_summary_data(pol))
        output_list.append(tabulate(no_match_data, headers, tablefmt="plain"))
    if len(has_matching) > 0:
        output_list.append(
            "\nPolicies WITH matching fingerprints in last query:"
        )
        data = []
        for pol in has_matching:
            data.append(policy_summary_data(pol))
        data.sort(key=lambda x: [x[3], x[1]])
        output_list.append(tabulate(data, headers, tablefmt="plain"))
    return "\n".join(output_list)


def policies_summary_output(
    policies: List[Dict],
    time: Tuple[float, float] = None,
    get_deviations_count: bool = False,
    suppress_msg=False,
    dev_name_or_uid=None,
):
    output_list = []
    if get_deviations_count:
        headers = [
            "UID",
            "NAME",
            "STATUS",
            "TYPE",
            "CREATE_TIME",
            "DEVIATIONS_(UNIQ/TOT)",
        ]
    else:
        headers = ["UID", "NAME", "STATUS", "TYPE", "CREATE_TIME"]
    data = []
    deviation_counts = {}
    if get_deviations_count:
        deviation_counts = get_deviation_counts(
            policies, time, suppress_msg, dev_name_or_uid
        )
    for policy in policies:
        data.append(
            policy_summary_data(policy, deviation_counts, get_deviations_count)
        )
    data.sort(key=lambda x: [x[3], x[1]])
    output_list.append(tabulate(data, headers, tablefmt="plain"))
    return "\n".join(output_list)


def policy_summary_data(
    policy: Dict, deviations_counts={}, get_deviations_count=False
):
    uid = policy[lib.METADATA_FIELD].get(lib.METADATA_UID_FIELD)
    status = policy[lib.SPEC_FIELD].get(lib.ENABLED_FIELD, True)
    mode = policy[lib.SPEC_FIELD].get(lib.POL_MODE_FIELD, lib.POL_MODE_ENFORCE)
    if status is False and uid:
        status = "Disabled"
    elif status is False and not uid:
        status = "Not Applied & Disabled"
    elif status and mode == lib.POL_MODE_ENFORCE and uid:
        status = "Enforcing"
    elif status and mode == lib.POL_MODE_AUDIT and uid:
        status = "Auditing"
    else:
        status = "Not Applied"
    if not uid:
        uid = "N/A"
    create_time = policy[lib.METADATA_FIELD].get(lib.METADATA_CREATE_TIME)
    if create_time:
        try:
            create_time = (
                zulu.parse(create_time).format("YYYY-MM-ddTHH:mm:ss") + "Z"
            )
        except Exception:
            pass
    else:
        create_time = "N/A"
    rv = [
        uid,
        policy[lib.METADATA_FIELD][lib.NAME_FIELD],
        status,
        policy[lib.METADATA_FIELD][lib.TYPE_FIELD],
        create_time,
    ]
    if get_deviations_count:
        counts = deviations_counts.get(uid)
        if not counts:
            rv.append("0/0")
        else:
            rv.append(f"{len(counts[0])}/{counts[1]}")
    return rv


def get_policy_by_uid(
    uid: str, policies: Optional[List[Dict]] = None
) -> Optional[Dict]:
    from spyctl.api import get_policies
    from spyctl.config.configs import get_current_context
    from spyctl.filter_resource import filter_obj

    ctx = get_current_context()
    if not policies:
        policies = get_policies(*ctx.get_api_data())
    policies = filter_obj(
        policies,
        [
            [lib.METADATA_FIELD, lib.METADATA_UID_FIELD],
        ],
        uid,
    )
    if not policies:
        return None
    return policies[0]


def get_deviation_counts(
    policies: List[Dict],
    time: Tuple[float, float],
    suppress_msg=False,
    dev_name_or_uid=None,
) -> Dict:
    if not suppress_msg:
        cli.try_log(
            f"Getting policy deviations from {lib.epoch_to_zulu(time[0])} to"
            f" {lib.epoch_to_zulu(time[1])}"
        )
    rv: Dict[str, List[Set, int]] = {}
    ctx = cfg.get_current_context()
    policy_uids = [
        policy[lib.METADATA_FIELD].get(lib.METADATA_UID_FIELD)
        for policy in policies
    ]
    pipeline = _af.Deviations.generate_count_pipeline(dev_name_or_uid)
    checksums = set()
    for count_obj in api.get_deviations(
        *ctx.get_api_data(),
        policy_uids,
        time,
        pipeline,
    ):
        uid = count_obj["policy_uid"]
        checksums.update(count_obj["counts"])
        if uid not in rv:
            rv[uid] = [set(), 0]
        rv[uid][0].update(count_obj["counts"])
        rv[uid][1] += sum(count_obj["counts"].values())
    return rv


def filter_policies_by_name_or_uid(policies, name_or_uid):
    if name_or_uid:
        policies = filt.filter_obj(
            policies,
            [
                [lib.METADATA_FIELD, lib.NAME_FIELD],
                [lib.METADATA_FIELD, lib.METADATA_UID_FIELD],
            ],
            name_or_uid,
        )
    else:
        return policies



import os
from contextlib import redirect_stderr
import time

import spyctl.spyctl_lib as lib


def test_label_only_key_inp():
    test_str1 = "env"
    test_output1 = {"env": "*"}
    result1 = lib.label_input_to_dict(test_str1)
    assert result1 == test_output1

    test_str2 = "env, tier"
    test_output2 = {"env": "*", "tier": "*"}
    result2 = lib.label_input_to_dict(test_str2)
    assert result2 == test_output2


def test_label_set_inp():
    test_str1 = "env in (stage, prod)"
    test_output1 = {"env": ["stage", "prod"]}
    result1 = lib.label_input_to_dict(test_str1)
    assert result1 == test_output1

    test_str2 = "env in (stage, prod), tier"
    test_output2 = {"env": ["stage", "prod"], "tier": "*"}
    result2 = lib.label_input_to_dict(test_str2)
    assert result2 == test_output2


def test_label_eq_inp():
    test_str1 = "env=stage"
    test_output1 = {"env": "stage"}
    result1 = lib.label_input_to_dict(test_str1)
    assert result1 == test_output1

    test_str2 = "env=prod, tier=frontend"
    test_output2 = {"env": "prod", "tier": "frontend"}
    result2 = lib.label_input_to_dict(test_str2)
    assert result2 == test_output2

    test_str3 = "env=prod, tier=frontend, app"
    test_output3 = {"env": "prod", "tier": "frontend", "app": "*"}
    result3 = lib.label_input_to_dict(test_str3)
    assert result3 == test_output3

    test_str4 = "env=prod,tier=frontend,app"
    test_output4 = {"env": "prod", "tier": "frontend", "app": "*"}
    result4 = lib.label_input_to_dict(test_str4)
    assert result4 == test_output4


def test_label_list_inp():
    test_strs1 = ["env=stage"]
    test_output1 = {"env": "stage"}
    result1 = lib.label_input_to_dict(test_strs1)
    assert result1 == test_output1

    test_strs2 = ["env=prod, tier=frontend", "app in (web,db)"]
    test_output2 = {"env": "prod", "tier": "frontend", "app": ["web", "db"]}
    result2 = lib.label_input_to_dict(test_strs2)
    assert result2 == test_output2


def test_label_dict_inp():
    test_dict = {"env": "stage", "tier": "frontend", "app": ["web", "db"]}
    result = lib.label_input_to_dict(test_dict)
    assert result == test_dict


def test_notin_inp():
    err_log = "err.log"
    with open(err_log, "w") as stderr, redirect_stderr(stderr):
        test_str = "env notin (stage, prod)"
        result = lib.label_input_to_dict(test_str)
        assert result is None
    with open(err_log, "r") as stderr:
        err = stderr.readline()
        assert "notin not supported" in err

    os.remove(err_log)

    with open(err_log, "w") as stderr, redirect_stderr(stderr):
        test_str = "env notin (stage, prod), app in (web)"
        result = lib.label_input_to_dict(test_str)
        assert result is None
    with open(err_log, "r") as stderr:
        err = stderr.readline()
        assert "notin not supported" in err

    os.remove(err_log)


def test_timeinp():
    assert lib.time_inp("1677522138", cap_one_day=False) == 1677522138
    tests = (
        ("50s", lambda t: t - 50),
        ("4m", lambda t: t - 4 * 60),
        ("7hr", lambda t: t - 7 * 60 * 60),
        ("1d", lambda t: t - 24 * 60 * 60),
        ("3w", lambda t: t - 3 * 7 * 24 * 60 * 60),
        # ("Mon Feb 27 13:43:39 CST 2023", lambda _: 1677527019),
        ("2023/02/26 23:34:32", lambda _: 1677454472),
        ("01/26/2023", lambda _: 1674691200),
    )
    for inp, out in tests:
        t = time.time()
        assert abs(lib.time_inp(inp, cap_one_day=False) - out(t)) < 1
    capped_inp = lib.time_inp("09/16/2020", cap_one_day=True)
    one_day_ago = time.time() - 24 * 60 * 60
    assert abs(capped_inp - one_day_ago) < 1



from __future__ import annotations

import time
from argparse import *
from argparse import _SubParsersAction
from datetime import datetime
from typing import Dict

import dateutil.parser as dateparser

from spyctl.resources.fingerprints import FPRINT_TYPE_CONT, FPRINT_TYPE_SVC
from spyctl.spyctl_lib import *

OUTPUT_YAML = "yaml"
OUTPUT_JSON = "json"
OUTPUT_SUMMARY = "summary"
OUTPUT_DEFAULT = OUTPUT_YAML
OUTPUT_CHOICES = [OUTPUT_JSON, OUTPUT_YAML]
DEFAULT_API_URL = "https://api.prod.spyderbat.com"
SP_POL_NAMES_AND_ALIASES = [
    "spyderbat-policy",
    "spyderbat-policies",
    "spy-pol",
    "spol",
    "sp",
]
UPLOAD_NAMES_AND_ALIASES = ["upload", "push"]
DELETE_NAMES_AND_ALIASES = ["delete", "del"]


class CustomHelpFormatter(RawDescriptionHelpFormatter):
    def _format_action_invocation(self, action):
        # changes "-p PODS, --pods PODS" tp "-p, --pods PODS"
        if action.option_strings:
            default = self._get_default_metavar_for_optional(action)
            args_string = self._format_args(action, default)
            return ", ".join(action.option_strings) + " " + args_string
        # changes subcommands "{get,merge,compare}" to "get, merge, compare"
        # and removes aliases from the list
        elif action.choices is not None and action.metavar is None:
            choice_strs = []
            try:
                obj_set = set()
                for choice, obj in action.choices.items():
                    if obj not in obj_set:
                        obj_set.add(obj)
                        choice_strs.append(str(choice))
            except AttributeError:
                choice_strs = [str(choice) for choice in action.choices]
            return ", ".join(choice_strs)
        else:
            return super()._format_action_invocation(action)

    def _format_args(self, action: Action, default_metavar: str) -> str:
        # changes "[files [files ...]]" to "[files ...]"
        if action.nargs == ZERO_OR_MORE:
            get_metavar = self._metavar_formatter(action, default_metavar)
            return "[%s ...]" % get_metavar(1)
        # changes "files [files ...]" to "files [...]"
        elif action.nargs == ONE_OR_MORE:
            get_metavar = self._metavar_formatter(action, default_metavar)
            return "%s [...]" % get_metavar(1)
        return super()._format_args(action, default_metavar)

    def _metavar_formatter(self, action: Action, default_metavar: str):
        # removes aliases
        if action.metavar is None and action.choices is not None:
            choice_strs = []
            try:
                obj_set = set()
                for choice, obj in action.choices.items():
                    if obj not in obj_set:
                        obj_set.add(obj)
                        choice_strs.append(str(choice))
            except AttributeError:
                choice_strs = [str(choice) for choice in action.choices]
            result = "{%s}" % ",".join(choice_strs)
            return lambda size: (result,) * size
        return super()._metavar_formatter(action, default_metavar)


# create a keyvalue class
# https://www.geeksforgeeks.org/python-key-value-pair-using-argparse/
class keyvalue(Action):
    # Constructor calling
    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, dict())

        for value in values:
            # split it into key and value
            key, value = value.split("=")
            # assign into dictionary
            getattr(namespace, self.dest)[key] = value


def fmt(prog):
    return CustomHelpFormatter(prog)


def time_inp(time_str: str) -> int:
    past_seconds = 0
    try:
        try:
            past_seconds = int(time_str) * 60
        except ValueError:
            if time_str.endswith(("s", "sc")):
                past_seconds = int(time_str[:-1])
            elif time_str.endswith(("m", "mn")):
                past_seconds = int(time_str[:-1]) * 60
            elif time_str.endswith(("h", "hr")):
                past_seconds = int(time_str[:-1]) * 60 * 60
            elif time_str.endswith(("d", "dy")):
                past_seconds = int(time_str[:-1]) * 60 * 60 * 24
            elif time_str.endswith(("w", "wk")):
                past_seconds = int(time_str[:-1]) * 60 * 60 * 24 * 7
            else:
                date = dateparser.parse(time_str)
                diff = datetime.now() - date
                past_seconds = diff.total_seconds()
    except (ValueError, dateparser.ParserError):
        raise ValueError("invalid time input (see global help menu)") from None
    if past_seconds < 0:
        raise ValueError("time must be in the past")
    return int(time.time() - past_seconds)


command_names = {}


def name_and_aliases(names):
    command_names[names[0]] = names
    if len(names) == 0:
        return dict(name=names[0])
    return dict(name=names[0], aliases=names[1:])


def get_names(command):
    return command_names.get(command, command)


def add_output_arg(parser, output_choices=OUTPUT_CHOICES):
    parser.add_argument(
        "-o",
        "--output",
        choices=output_choices,
        type=output_argument_helper,
        default=OUTPUT_DEFAULT,
    )
    parser.add_argument(
        "-f",
        "--filter",
        help="filter output objects by string",
        action="append",
        default=[],
    )


def add_policy_file_arg(parser: ArgumentParser):
    parser.add_argument(
        "file",
        help="file containing spyderbat policy",
        type=FileType("r"),
        nargs="?",
    )


def output_argument_helper(input: str) -> str:
    i = input.lower()
    if i.startswith("y"):
        return OUTPUT_YAML
    elif i.startswith("j"):
        return OUTPUT_JSON
    elif i.startswith("s"):
        return OUTPUT_SUMMARY
    else:
        return i


def add_time_arg(parser):
    times = parser.add_mutually_exclusive_group()
    times.add_argument(
        "-t", "--time", help="collect data at this time", type=time_inp
    )
    times.add_argument(
        "-r",
        "--time-range",
        help="collect data between two times",
        action="extend",
        nargs=2,
        type=time_inp,
    )
    times.add_argument(
        "-w",
        "--within",
        help="collect data from this time until now",
        metavar="TIME",
        type=time_inp,
    )


def parse_args():
    desc = "a tool to help manage Spyderbat fingerprints and policies"
    epilog = "see TUTORIAL.md for an explanation of usage"
    parser = ArgumentParser(
        description=desc, epilog=epilog, formatter_class=fmt
    )
    parser.add_argument(
        "--deployment",
        help="name of deployment to use from user configuration file",
        dest="selected_deployment",
    )
    parser.add_argument(
        "-y",
        "--yes",
        help="automatic yes to prompts; assume 'yes' as answer to all prompts"
        " and run non-interactively.",
        action="store_true",
    )
    subs = parser.add_subparsers(
        title="subcommands", dest="subcommand", required=True
    )
    make_compare(subs)
    make_configure(subs)
    make_create(subs)
    make_delete(subs)
    make_get(subs)
    make_manage(subs)
    make_merge(subs)
    make_upload(subs)
    return parser.parse_args()


def make_configure(subs: _SubParsersAction[ArgumentParser]):
    desc = "configure API information"
    names = ["configure", "conf", "config"]
    configure = subs.add_parser(
        **name_and_aliases(names), description=desc, formatter_class=fmt
    )
    configure_subs = configure.add_subparsers(
        title="configure commands", dest="config_cmd"
    )
    make_configure_add(configure_subs)
    make_configure_update(configure_subs)
    make_configure_setdefault(configure_subs)
    make_configure_delete(configure_subs)
    make_configure_show(configure_subs)


def make_configure_add(configure_subs: _SubParsersAction[ArgumentParser]):
    desc = "add Spyderbat API deployment to spyctl user configuration"
    configure_add = configure_subs.add_parser(
        "add", description=desc, formatter_class=fmt
    )
    configure_add.add_argument(
        "deployment_name", help="the name of the deployment to add"
    )
    configure_add.add_argument("-k", "--api_key", metavar="KEY", required=True)
    configure_add.add_argument(
        "-u",
        "--api_url",
        help=f"Default: {DEFAULT_API_URL}",
        default=DEFAULT_API_URL,
        metavar="URL",
    )
    configure_add.add_argument(
        "-o", "--org", metavar="NAME_OR_UID", required=True
    )
    configure_add.add_argument(
        "-D",
        "--set-default",
        action="store_true",
        help="set config as default",
    )


def make_configure_update(configure_subs: _SubParsersAction[ArgumentParser]):
    desc = (
        "update a specific Spyderbat API deployment in spyctl user"
        " configuration"
    )
    configure_update = configure_subs.add_parser(
        "update", description=desc, formatter_class=fmt
    )
    configure_update.add_argument(
        "deployment_name", help="the name of the deployment to update"
    )
    configure_update.add_argument("-k", "--api_key", metavar="KEY")
    configure_update.add_argument("-u", "--api_url", metavar="URL")
    configure_update.add_argument("-o", "--org", metavar="NAME_OR_UID")
    configure_update.add_argument(
        "-D",
        "--set-default",
        action="store_true",
        help="set deployment as default",
    )


def make_configure_setdefault(
    configure_subs: _SubParsersAction[ArgumentParser],
):
    desc = (
        "set a Spyderbat API deployment as default in spyctl user",
        " configuration",
    )
    configure_default = configure_subs.add_parser(
        "default", description=desc, formatter_class=fmt
    )
    configure_default.add_argument(
        "deployment_name", help="the name of the deployment to set as default"
    )


def make_configure_delete(configure_subs: _SubParsersAction[ArgumentParser]):
    desc = "delete a Spyderbat API deployment from spyctl user configuration"
    configure_delete = configure_subs.add_parser(
        "delete", description=desc, formatter_class=fmt
    )
    configure_delete.add_argument(
        "deployment_name", help="the name of the deployment to set as default"
    )


def make_configure_show(configure_subs: _SubParsersAction[ArgumentParser]):
    desc = "show the user configuration file"
    configure_show = configure_subs.add_parser(
        "show", description=desc, formatter_class=fmt
    )


def make_delete(subs: _SubParsersAction[ArgumentParser]):
    desc = "Delete Spyderbat objects"
    names = DELETE_NAMES_AND_ALIASES
    delete = subs.add_parser(
        **name_and_aliases(DELETE_NAMES_AND_ALIASES),
        description=desc,
        formatter_class=fmt,
    )
    delete_subs = delete.add_subparsers(
        title="delete commands", dest="delete_cmd"
    )
    make_delete_policy(SP_POL_NAMES_AND_ALIASES, delete_subs)


def make_delete_policy(names, delete_subs: _SubParsersAction[ArgumentParser]):
    desc = "Delete Spyderbat policy via API call"
    delete_policy = delete_subs.add_parser(
        **name_and_aliases(names), description=desc, formatter_class=fmt
    )
    del_pol_grp = delete_policy.add_mutually_exclusive_group(required=True)
    del_pol_grp.add_argument(
        "-u", "--uid", help="UID of Spyderbat Policy object"
    )
    del_pol_grp.add_argument(
        "-P",
        "--policy",
        help="Spyderbat Policy with uid in metadata field",
        dest="policy_file",
    )


def make_manage(subs: _SubParsersAction[ArgumentParser]):
    desc = "Edit Spyderbat objects"
    manage = subs.add_parser("manage", description=desc, formatter_class=fmt)
    manage_subs = manage.add_subparsers(
        title="manage commands", dest="manage_cmd"
    )
    make_manage_policy(manage_subs)


def make_manage_policy(manage_subs: _SubParsersAction[ArgumentParser]):
    desc = "Manage Spyderbat policy objects"
    names = SP_POL_NAMES_AND_ALIASES
    manage_policy = manage_subs.add_parser(
        **name_and_aliases(names), description=desc, formatter_class=fmt
    )
    man_pol_subs = manage_policy.add_subparsers(
        title="policy management commands", dest="pol_cmd"
    )
    make_manage_policy_add_response(man_pol_subs)
    make_manage_policy_enable(man_pol_subs)
    make_manage_policy_disable(man_pol_subs)
    make_upload_policy(UPLOAD_NAMES_AND_ALIASES, man_pol_subs)
    make_delete_policy(DELETE_NAMES_AND_ALIASES, man_pol_subs)


def make_manage_policy_add_response(
    man_pol_subs: _SubParsersAction[ArgumentParser],
):
    desc = "Add a response action to a Spyderbat policy object"
    man_pol_add_resp = man_pol_subs.add_parser(
        "add-response", description=desc, formatter_class=fmt
    )
    add_policy_file_arg(man_pol_add_resp)
    man_pol_add_resp.add_argument(
        "-s",
        "--severity",
        choices=ALLOWED_SEVERITIES,
        help="severity of spyderbat flag for matching policy deviation",
    )
    man_pol_add_resp.add_argument(
        "-a",
        "--action",
        choices=ALLOWED_ACTIONS,
        help="the type of response action to be taken",
        required=True,
    )
    man_pol_add_resp.add_argument(
        "-p",
        "--pod-labels",
        metavar="POD_LABEL",
        help="key=value pairs separated by space for use in matching against"
        " k8s pod labels (ex. 'env=prod app=website)'",
        nargs="*",
        action=keyvalue,
    )
    man_pol_add_resp.add_argument(
        "-n",
        "--namespace-labels",
        metavar="NAMESPACE_LABEL",
        help="key=value pairs separated by space for use in matching against"
        " k8s namespace labels (ex. 'env=prod app=website')",
        nargs="*",
        action=keyvalue,
    )
    add_output_arg(man_pol_add_resp)
    webhook_group_desc = f"options required when action is '{ACTION_WEBHOOK}'"
    webhook_group = man_pol_add_resp.add_argument_group(
        "webhook arguments", webhook_group_desc
    )
    webhook_group.add_argument(
        "-t",
        "--template",
        choices=ALLOWED_TEMPLATES,
        help="the structure of the webhook output",
    )
    webhook_group.add_argument(
        "-u",
        "--url",
        help="url destination for the webhook",
    )


def make_manage_policy_enable(
    man_pol_subs: _SubParsersAction[ArgumentParser],
):
    desc = (
        "Enable a Spyderbat policy object\n  note: policies are enabled by"
        " default"
    )
    man_pol_enable = man_pol_subs.add_parser(
        "enable", description=desc, formatter_class=fmt
    )
    add_policy_file_arg(man_pol_enable)
    add_output_arg(man_pol_enable)


def make_manage_policy_disable(
    man_pol_subs: _SubParsersAction[ArgumentParser],
):
    desc = (
        "Disable a Spyderbat policy object\n  note: this adds an"
        f" '{ENABLED_FIELD}: false' field to the policy object."
        " The policy must be uploaded for this to take effect"
    )
    man_pol_disable = man_pol_subs.add_parser(
        "disable", description=desc, formatter_class=fmt
    )
    add_policy_file_arg(man_pol_disable)
    add_output_arg(man_pol_disable)


def make_create(subs: _SubParsersAction[ArgumentParser]):
    desc = "create YAML objects for use in Spyderbat"
    create = subs.add_parser("create", description=desc, formatter_class=fmt)
    create_subs = create.add_subparsers(
        title="output objects", dest="create_target"
    )
    make_create_spyderbat_policy(create_subs)


def make_create_spyderbat_policy(
    create_subs: _SubParsersAction[ArgumentParser],
):
    names = SP_POL_NAMES_AND_ALIASES
    create_spyderbat_policy = create_subs.add_parser(
        **name_and_aliases(names), formatter_class=fmt
    )
    create_spyderbat_policy.add_argument(
        "-F",
        "--fingerprints",
        help="fingerprint(s) to build a policy from",
        nargs="?",
        const="-",
    )
    create_spyderbat_policy.add_argument(
        "type",
        choices=[POL_TYPE_CONT, POL_TYPE_SVC],
        help="the type of policy to create",
    )
    add_output_arg(create_spyderbat_policy)


def make_get(subs: _SubParsersAction[ArgumentParser]):
    desc = "get different types of information and objects"
    names = ["get", "pull"]
    get = subs.add_parser(
        **name_and_aliases(names), description=desc, formatter_class=fmt
    )
    get_subs = get.add_subparsers(title="targets", dest="get_target")
    make_get_cluster(get_subs)
    make_get_namespace(get_subs)
    make_get_machine(get_subs)
    make_get_pod(get_subs)
    make_get_fingerprint(get_subs)
    make_get_policy(get_subs)


def make_get_cluster(get_subs: _SubParsersAction[ArgumentParser]):
    names = ["clusters", "clust", "clusts", "cluster"]
    get_cluster = get_subs.add_parser(
        **name_and_aliases(names), formatter_class=fmt
    )
    add_time_arg(get_cluster)
    add_output_arg(get_cluster)


def make_get_namespace(get_subs: _SubParsersAction[ArgumentParser]):
    names = ["namespaces", "name", "names", "namesp", "namesps", "namespace"]
    get_namespace = get_subs.add_parser(
        **name_and_aliases(names), formatter_class=fmt
    )
    get_namespace.add_argument(
        "-c", "--clusters", nargs="?", const="-", required=True
    )
    add_time_arg(get_namespace)
    add_output_arg(get_namespace)


def make_get_machine(get_subs: _SubParsersAction[ArgumentParser]):
    names = ["machines", "mach", "machs", "machine"]
    get_machine = get_subs.add_parser(
        **name_and_aliases(names), formatter_class=fmt
    )
    get_machine.add_argument("-c", "--clusters", nargs="?", const="-")
    add_time_arg(get_machine)
    add_output_arg(get_machine)


def make_get_pod(get_subs: _SubParsersAction[ArgumentParser]):
    names = ["pods", "pod"]
    get_pod = get_subs.add_parser(
        **name_and_aliases(names), formatter_class=fmt
    )
    get_pod.add_argument("-c", "--clusters", nargs="?", const="-")
    get_pod.add_argument("-m", "--machines", nargs="?", const="-")
    get_pod.add_argument("-n", "--namespaces", nargs="?", const="-")
    add_time_arg(get_pod)
    add_output_arg(get_pod)


def make_get_fingerprint(get_subs: _SubParsersAction[ArgumentParser]):
    names = ["fingerprints", "print", "prints", "fingerprint"]
    get_fingerprint = get_subs.add_parser(
        **name_and_aliases(names), formatter_class=fmt
    )
    get_fingerprint.add_argument(
        "type",
        choices=[FPRINT_TYPE_CONT, FPRINT_TYPE_SVC],
        help="the type of fingerprints to get",
    )
    selector_group = get_fingerprint.add_mutually_exclusive_group()
    selector_group.add_argument("-c", "--clusters", nargs="?", const="-")
    selector_group.add_argument("-m", "--machines", nargs="?", const="-")
    selector_group.add_argument("-p", "--pods", nargs="?", const="-")
    add_time_arg(get_fingerprint)
    output_choices = OUTPUT_CHOICES + [OUTPUT_SUMMARY]
    add_output_arg(get_fingerprint, output_choices)


def make_get_policy(get_subs: _SubParsersAction[ArgumentParser]):
    names = SP_POL_NAMES_AND_ALIASES
    get_policy = get_subs.add_parser(
        **name_and_aliases(names), formatter_class=fmt
    )
    get_policy.add_argument(
        "type",
        choices=[POL_TYPE_CONT, POL_TYPE_SVC],
        help="the type of policy to get",
    )
    pol_uid_grp = get_policy.add_mutually_exclusive_group()
    pol_uid_grp.add_argument(
        "-u", "--uid", help="UID of Spyderbat Policy object"
    )
    pol_uid_grp.add_argument(
        "-P",
        "--policy",
        help="Spyderbat Policy with uid in metadata field",
        dest="policy_file",
    )
    add_output_arg(get_policy)


def make_compare(subs: _SubParsersAction[ArgumentParser]):
    desc = "compare a set of fingerprints"
    names = ["compare", "diff", "comp"]
    compare = subs.add_parser(
        **name_and_aliases(names), description=desc, formatter_class=fmt
    )
    compare.add_argument(
        "files",
        help="fingerprint files to compare",
        type=FileType("r"),
        nargs="*",
    )


def make_merge(subs: _SubParsersAction[ArgumentParser]):
    desc = "merge a set of fingerprints"
    names = ["merge", "combine"]
    merge = subs.add_parser(
        **name_and_aliases(names), description=desc, formatter_class=fmt
    )
    merge.add_argument(
        "files",
        help="fingerprint files to merge",
        type=FileType("r"),
        nargs="*",
    )
    add_output_arg(merge)


def make_upload(subs: _SubParsersAction[ArgumentParser]):
    desc = "Upload an object to Spyderbat's backend"
    names = UPLOAD_NAMES_AND_ALIASES
    upload = subs.add_parser(
        **name_and_aliases(names), description=desc, formatter_class=fmt
    )
    upload_subs = upload.add_subparsers(
        title="upload commands", description=desc, dest="upload_cmd"
    )
    make_upload_policy(SP_POL_NAMES_AND_ALIASES, upload_subs)


def make_upload_policy(names, upload_subs: _SubParsersAction[ArgumentParser]):
    upload_policy = upload_subs.add_parser(
        **name_and_aliases(names), formatter_class=fmt
    )
    upload_policy.add_argument(
        "-u", "--uid", help="uid of a previously-uploaded policy to update"
    )
    add_policy_file_arg(upload_policy)



import spyctl.spyctl_lib as s_lib
import spyctl.config.configs as cfg
from fastapi import HTTPException


def flush_spyctl_log_messages() -> str:
    messages = s_lib.flush_log_var() + "\n" + s_lib.flush_err_var()
    return messages


def generate_spyctl_context(
    org_uid: str, api_key: str, api_url: str
) -> cfg.Context:
    if not org_uid or not api_key or not api_url:
        raise HTTPException(
            status_code=400,
            detail="Bad Request. Missing org_uid, api_key, and/or api_url",
        )
    spyctl_ctx = cfg.create_temp_secret_and_context(org_uid, api_key, api_url)
    return spyctl_ctx
